# Copyright (c) 2017 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import annotations
from typing import List, Optional, Tuple, TYPE_CHECKING

import numpy
from numpy import uint32
from numpy.typing import NDArray

from pacman.model.graphs.common import Slice
from pacman.model.placements import Placement

from spinn_front_end_common.utilities.helpful_functions import (
    locate_memory_region_for_placement)
from spinn_front_end_common.utilities.constants import BYTES_PER_WORD
from spinn_front_end_common.interface.buffer_management import BufferManager

from spynnaker.pyNN.data import SpynnakerDataView
from spynnaker.pyNN.models.neuron.synapse_dynamics import (
    AbstractSynapseDynamicsStructural)
from spynnaker.pyNN.models.neural_projections.connectors import (
    AbstractGenerateConnectorOnHost)
from spynnaker.pyNN.types import WeightScales

from .generator_data import GeneratorData
from .synapse_io import read_all_synapses, convert_to_connections, get_synapses

if TYPE_CHECKING:
    from spynnaker.pyNN.models.neural_projections import (
        ProjectionApplicationEdge, SynapseInformation)
    from spynnaker.pyNN.models.neuron.synaptic_matrices import AppKeyInfo
    from .master_pop_table import MasterPopTableAsBinarySearch


class SynapticMatrixApp(object):
    """
    The synaptic matrix (and delay matrix if applicable) for an incoming
    application edge.
    """

    __slots__ = (
        # The synaptic info that these matrices are for
        "__synapse_info",
        # The application edge that these matrices are for
        "__app_edge",
        # The number of synapse types incoming
        "__n_synapse_types",
        # The ID of the synaptic matrix region
        "__synaptic_matrix_region",
        # The maximum row length of delayed and undelayed matrices
        "__max_row_info",
        # The maximum summed size of the synaptic matrices
        "__all_syn_block_sz",
        # The application-level key information for the incoming edge
        "__app_key_info",
        # The application-level key information for the incoming delay edge
        "__delay_app_key_info",
        # The weight scaling used by each synapse type
        "__weight_scales",
        # The expected size in bytes of a synaptic matrix
        "__matrix_size",
        # The expected size in bytes of a delayed synaptic matrix
        "__delay_matrix_size",
        # The offset of the undelayed synaptic matrix in the region
        "__syn_mat_offset",
        # The offset of the delayed synaptic matrix in the region
        "__delay_syn_mat_offset",
        # The index of the synaptic matrix within the master population table
        "__index",
        # The number of bits to use for neuron IDs
        "__max_atoms_per_core",
        # The download index for the undelayed synaptic matrix
        "__download_index",
        # The download index for the delayed synaptic matrix
        "__download_delay_index")

    def __init__(
            self, synapse_info: SynapseInformation,
            app_edge: ProjectionApplicationEdge, n_synapse_types: int,
            synaptic_matrix_region: int, max_atoms_per_core: int,
            all_syn_block_sz: int, app_key_info: Optional[AppKeyInfo],
            delay_app_key_info: Optional[AppKeyInfo],
            weight_scales: WeightScales):
        """
        :param synapse_info:
            The projection synapse information
        :param app_edge:
            The projection application edge
        :param n_synapse_types: The number of synapse types accepted
        :param synaptic_matrix_region:
            The region where synaptic matrices are stored
        :param all_syn_block_sz:
            The space available for all synaptic matrices on the core
        :param app_key_info:
            Application-level routing key information for undelayed vertices
        :param delay_app_key_info:
            Application-level routing key information for delayed vertices
        :param weight_scales:
            Weight scale for each synapse edge
        """
        self.__synapse_info = synapse_info
        self.__app_edge = app_edge
        self.__n_synapse_types = n_synapse_types
        self.__synaptic_matrix_region = synaptic_matrix_region
        self.__max_atoms_per_core = max_atoms_per_core

        # Calculate the max row info for this edge
        self.__max_row_info = self.__app_edge.post_vertex.get_max_row_info(
            synapse_info, max_atoms_per_core, app_edge)

        self.__all_syn_block_sz = all_syn_block_sz
        self.__app_key_info = app_key_info
        self.__delay_app_key_info = delay_app_key_info
        self.__weight_scales = weight_scales

        self.__matrix_size = (
            self.__app_edge.pre_vertex.n_atoms *
            self.__max_row_info.undelayed_max_bytes)
        self.__delay_matrix_size = (
            self.__app_edge.pre_vertex.n_atoms *
            self.__app_edge.n_delay_stages *
            self.__max_row_info.delayed_max_bytes)

        # These are computed during synaptic generation
        self.__syn_mat_offset: Optional[int] = None
        self.__delay_syn_mat_offset: Optional[int] = None
        self.__index: Optional[int] = None

        self.__download_index: Optional[int] = None
        self.__download_delay_index: Optional[int] = None

    @property
    def gen_size(self) -> int:
        """
        Size of a block.
        """
        max_row_length = max(
            self.__max_row_info.undelayed_max_bytes,
            self.__max_row_info.delayed_max_bytes)
        return (max_row_length * self.__app_edge.pre_vertex.n_atoms *
                (self.__app_edge.n_delay_stages + 1))

    def reserve_matrices(
            self, block_addr: int,
            pop_table: MasterPopTableAsBinarySearch) -> int:
        """
        Allocate the master pop table entries for the blocks.

        :param block_addr: Where the allocation can start from
        :param pop_table: The master population table
        :return: Where the next allocation can start from
        """
        block_addr = self.__reserve_app_matrix(block_addr, pop_table)
        block_addr = self.__reserve_delay_app_matrix(block_addr, pop_table)
        return block_addr

    def __reserve_app_matrix(
            self, block_addr: int,
            pop_table: MasterPopTableAsBinarySearch) -> int:
        """
        Reserve space for the matrix in the master pop table.

        :param block_addr:
            The address in the synaptic matrix region to start writing at
        :param pop_table:
            The master population table
        :return: The updated block address
        """
        # If there is no routing info, don't write anything
        if self.__app_key_info is None:
            return block_addr

        # If we have routing info but no synapses, write an invalid entry
        if self.__max_row_info.undelayed_max_n_synapses == 0:
            self.__index = pop_table.add_invalid_application_entry(
                self.__app_key_info.key_and_mask,
                self.__app_key_info.core_mask, self.__app_key_info.core_shift,
                self.__app_key_info.n_neurons,
                self.__app_key_info.n_colour_bits)
            return block_addr

        # Write a matrix for the whole application vertex
        block_addr = pop_table.get_next_allowed_address(block_addr)
        self.__index = pop_table.add_application_entry(
            block_addr,  self.__max_row_info.undelayed_max_words,
            self.__app_key_info.key_and_mask, self.__app_key_info.core_mask,
            self.__app_key_info.core_shift, self.__app_key_info.n_neurons,
            self.__app_key_info.n_colour_bits)
        self.__syn_mat_offset = block_addr
        block_addr = self.__next_addr(block_addr, self.__matrix_size)
        return block_addr

    def __reserve_delay_app_matrix(
            self, block_addr: int,
            pop_table: MasterPopTableAsBinarySearch) -> int:
        """
        Reserve space in the master pop table for a delayed matrix.

        :param block_addr:
            The address in the synaptic matrix region to start writing at
        :param pop_table:
            The master population table
        :return: The updated block address
        """
        # If there is no routing info, don't write anything
        if self.__delay_app_key_info is None:
            return block_addr

        # If we have routing info but no synapses, write an invalid entry
        if self.__max_row_info.delayed_max_n_synapses == 0:
            pop_table.add_invalid_application_entry(
                self.__delay_app_key_info.key_and_mask,
                self.__delay_app_key_info.core_mask,
                self.__delay_app_key_info.core_shift,
                self.__delay_app_key_info.n_neurons,
                self.__delay_app_key_info.n_colour_bits)
            return block_addr

        # Write a matrix for the whole application vertex
        block_addr = pop_table.get_next_allowed_address(block_addr)
        pop_table.add_application_entry(
            block_addr, self.__max_row_info.delayed_max_words,
            self.__delay_app_key_info.key_and_mask,
            self.__delay_app_key_info.core_mask,
            self.__delay_app_key_info.core_shift,
            self.__delay_app_key_info.n_neurons,
            self.__delay_app_key_info.n_colour_bits)
        self.__delay_syn_mat_offset = block_addr
        block_addr = self.__next_addr(block_addr, self.__delay_matrix_size)
        return block_addr

    def __next_addr(self, block_addr: int, size: int) -> int:
        """
        Get the next address after a block, checking it is in range.

        :param block_addr: The address of the start of the block
        :param size: The size of the block in bytes
        :return: The updated address
        :raises ValueError: If the updated address is out of range
        """
        next_addr = block_addr + size
        if next_addr > self.__all_syn_block_sz:
            raise ValueError(
                "Too much synaptic memory has been written: "
                f"{next_addr} of {self.__all_syn_block_sz} ")
        return next_addr

    def append_matrix(
            self, post_vertex_slice: Slice,
            data_to_write: List[NDArray[uint32]],
            block_addr: int) -> int:
        """
        Append a synaptic matrix from be written from host.

        :param post_vertex_slice:
            The slice of the post-vertex the matrix is for
        :param data_to_write: List to append the data to write to
        :param block_addr: The amount of data written so far
        :return: The amount of data written after this data has been written
        """
        row_data, delay_row_data = self.__get_row_data(post_vertex_slice)
        self.__update_connection_holders(
            row_data, delay_row_data, post_vertex_slice)
        if self.__syn_mat_offset is not None:
            block_addr = self.__get_padding(
                data_to_write, self.__syn_mat_offset, block_addr)
            data_to_write.append(row_data)
            block_addr += self.__matrix_size
        if self.__delay_syn_mat_offset is not None:
            block_addr = self.__get_padding(
                data_to_write, self.__delay_syn_mat_offset, block_addr)
            data_to_write.append(delay_row_data)
            block_addr += self.__delay_matrix_size
        return block_addr

    def __get_padding(
            self, data_to_write: List[NDArray[uint32]],
            expected_offset: int, block_addr: int) -> int:
        if expected_offset < block_addr:
            raise ValueError(
                "The block address is already beyond where is expected!:"
                f" {expected_offset} expected, {block_addr} found.")
        if expected_offset > block_addr:
            padding = (expected_offset - block_addr) // BYTES_PER_WORD
            data_to_write.append(numpy.zeros(padding, dtype=uint32))
            return block_addr + (padding * BYTES_PER_WORD)
        return block_addr

    def __get_row_data(
            self, post_vertex_slice: Slice) -> Tuple[NDArray, NDArray]:
        """
        Generate the row data for a synaptic matrix from the description.

        :return: The data and the delayed data
        """
        # Get the actual connections
        post_slices =\
            self.__app_edge.post_vertex.splitter.get_in_coming_slices()
        connector = self.__synapse_info.connector
        assert isinstance(connector, AbstractGenerateConnectorOnHost)
        connections = connector.create_synaptic_block(
            post_slices, post_vertex_slice,
            self.__synapse_info.synapse_type, self.__synapse_info)

        # Get the row data; note that we use the availability of the routing
        # keys to decide if we should actually generate any data; this is
        # because a single edge might have been filtered
        (row_data, delayed_row_data) = get_synapses(
            connections, self.__synapse_info, self.__app_edge.n_delay_stages,
            self.__n_synapse_types, self.__weight_scales, self.__app_edge,
            self.__max_row_info, self.__app_key_info is not None,
            self.__delay_app_key_info is not None, self.__max_atoms_per_core)

        # Set connections for structural plasticity
        if isinstance(self.__synapse_info.synapse_dynamics,
                      AbstractSynapseDynamicsStructural):
            self.__synapse_info.synapse_dynamics.set_connections(
                connections, post_vertex_slice, self.__app_edge,
                self.__synapse_info)
        if self.__app_edge.delay_edge is None and len(delayed_row_data) != 0:
            raise ValueError(
                "Found delayed source IDs but no delay "
                f"edge for {self.__app_edge.label}")

        return row_data, delayed_row_data

    def __update_connection_holders(
            self, data: NDArray[uint32], delayed_data: NDArray[uint32],
            post_vertex_slice: Slice) -> None:
        """
        Fill in connections in the connection holders as they are created.

        :param data: The row data created
        :param delayed_data: The delayed row data created
        """
        post_splitter = self.__app_edge.post_vertex.splitter
        post_vertex_max_delay_ticks = post_splitter.max_support_delay()
        for conn_holder in self.__synapse_info.pre_run_connection_holders:
            conn_holder.add_connections(read_all_synapses(
                data, delayed_data, self.__synapse_info,
                self.__n_synapse_types, self.__weight_scales,
                post_vertex_slice, self.__app_edge.pre_vertex.n_atoms,
                post_vertex_max_delay_ticks,
                self.__max_row_info, self.__max_atoms_per_core))

    def get_generator_data(self) -> GeneratorData:
        """
        Prepare to write a matrix using an on-chip generator.

        :return: The data to generate with
        """
        max_pre_atoms_per_core = min(
            self.__app_edge.pre_vertex.n_atoms,
            self.__app_edge.pre_vertex.get_max_atoms_per_core())
        return GeneratorData(
            self.__syn_mat_offset, self.__delay_syn_mat_offset,
            self.__app_edge, self.__synapse_info, self.__max_row_info,
            max_pre_atoms_per_core, self.__max_atoms_per_core)

    def read_generated_connection_holders(self, placement: Placement) -> None:
        """
        Read any pre-run connection holders after data has been generated.

        :param placement:
            Where the matrix is on the machine
        """
        if self.__synapse_info.pre_run_connection_holders:
            connections = self.get_connections(placement)
            if connections:
                conns = numpy.concatenate(connections)
                for holder in self.__synapse_info.pre_run_connection_holders:
                    holder.add_connections(conns)

    def get_connections(self, placement: Placement) -> List[NDArray]:
        """
        Read connections from an address on the machine.

        :param placement:
            Where the matrix is on the machine
        :return: A list of arrays of connections, each with dtype
            :py:const:`~.NUMPY_CONNECTORS_DTYPE`
        """
        connections = list()

        synapses_address: Optional[int] = None
        buffers: Optional[BufferManager] = None
        if (self.__download_index is None and
                self.__download_delay_index is None):
            synapses_address = locate_memory_region_for_placement(
                placement, self.__synaptic_matrix_region)
        else:
            buffers = SpynnakerDataView().get_buffer_manager()

        splitter = self.__app_edge.post_vertex.splitter
        vertex_slice = placement.vertex.vertex_slice
        if self.__syn_mat_offset is not None:
            if self.__download_index is not None:
                assert buffers is not None
                block, _ = buffers.get_download(
                    placement, self.__download_index)
            else:
                assert synapses_address is not None
                block = self.__get_block(placement, synapses_address)
            connections.append(convert_to_connections(
                self.__synapse_info, vertex_slice,
                self.__app_edge.pre_vertex.n_atoms,
                self.__max_row_info.undelayed_max_words,
                self.__n_synapse_types, self.__weight_scales, block, False,
                splitter.max_support_delay(), self.__max_atoms_per_core))
        if self.__delay_syn_mat_offset is not None:
            if self.__download_delay_index is not None:
                assert buffers is not None
                block, _ = buffers.get_download(
                    placement, self.__download_delay_index)
            else:
                assert synapses_address is not None
                block = self.__get_delayed_block(placement, synapses_address)
            connections.append(convert_to_connections(
                self.__synapse_info, vertex_slice,
                self.__app_edge.pre_vertex.n_atoms,
                self.__max_row_info.delayed_max_words,
                self.__n_synapse_types, self.__weight_scales, block, True,
                splitter.max_support_delay(), self.__max_atoms_per_core))

        return connections

    def __get_block(
            self, placement: Placement, synapses_address: int) -> bytes:
        """
        Get a block of data for undelayed synapses.

        :param placement: Where the matrix is on the machine
        :param synapses_address:
            The base address of the synaptic matrix region
        :return: The raw data from the synaptic matrix
        """
        assert self.__syn_mat_offset is not None
        address = self.__syn_mat_offset + synapses_address
        return SpynnakerDataView.read_memory(
            placement.x, placement.y, address, self.__matrix_size)

    def __get_delayed_block(
            self, placement: Placement, synapses_address: int) -> bytes:
        """
        Get a block of data for delayed synapses.

        :param placement:
            Where the matrix is on the machine
        :param synapses_address:
            The base address of the synaptic matrix region
        :return: The raw data from the delayed synaptic matrix
        """
        assert self.__delay_syn_mat_offset is not None
        address = self.__delay_syn_mat_offset + synapses_address
        return SpynnakerDataView.read_memory(
            placement.x, placement.y, address, self.__delay_matrix_size)

    def get_index(self) -> int:
        """
        :returns: The index in the master population table of the matrix.
        """
        if self.__index is None:
            raise RuntimeError("master pop table space not yet reserved")
        return self.__index

    def get_download_regions(
            self, placement: Placement,
            start_index: int) -> List[Tuple[int, int, int]]:
        """
        Get the data regions that should be downloaded when the simulation
        pauses.

        :param placement:
            The placement of the vertex
        :param start_index:
            The index to use for the first region to download

        :return: A list of tuples of (index, address, size) to download
        """
        if not self.__synapse_info.download_on_pause:
            return []
        synapses_address = locate_memory_region_for_placement(
            placement, self.__synaptic_matrix_region)
        regions = list()
        if self.__syn_mat_offset is not None:
            regions.append((start_index,
                            synapses_address + self.__syn_mat_offset,
                            self.__matrix_size))
            self.__download_index = start_index
            start_index += 1
        if self.__delay_syn_mat_offset is not None:
            regions.append((start_index,
                            synapses_address + self.__delay_syn_mat_offset,
                            self.__delay_matrix_size))
            self.__download_delay_index = start_index
        return regions



# Copyright (c) 2015 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Iterable
from pacman.model.placements import Placement


def make_missing_string(missing: Iterable[Placement]) -> str:
    """
    Unused likely to be removed.

    :param missing:
    :return: A String with the X, Y and P of each placement
    """
    return "; ".join(
        f"({placement.x}, {placement.y}, {placement.p})"
        for placement in missing)



# Copyright (c) 2015 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from spynnaker.pyNN.exceptions import SpynnakerException
from spynnaker.pyNN.models.defaults import (
    AbstractProvidesDefaults, default_initial_values)
from spynnaker.pyNN.models.neuron.implementations import ModelParameter


class HHCondExp(AbstractProvidesDefaults):
    """
    Single-compartment Hodgkin-Huxley model with exponentially decaying
    current input.

    .. warning::
        Not currently supported by the tool chain.
    """

    # noinspection PyPep8Naming
    @default_initial_values({"v", "gsyn_exc", "gsyn_inh"})
    def __init__(
            self, gbar_K: ModelParameter = 6.0, cm: ModelParameter = 0.2,
            e_rev_Na: ModelParameter = 50.0, tau_syn_E: ModelParameter = 0.2,
            tau_syn_I: ModelParameter = 2.0, i_offset: ModelParameter = 0.0,
            g_leak: ModelParameter = 0.01, e_rev_E: ModelParameter = 0.0,
            gbar_Na: ModelParameter = 20.0, e_rev_leak: ModelParameter = -65.0,
            e_rev_I: ModelParameter = -80, e_rev_K: ModelParameter = -90.0,
            v_offset: ModelParameter = -63, v: ModelParameter = -65.0,
            gsyn_exc: ModelParameter = 0.0, gsyn_inh: ModelParameter = 0.0):
        # pylint: disable=unused-argument,invalid-name
        """
        :param gbar_K:
        :param cm:
        :param e_rev_Na:
        :param tau_syn_E:
        :param tau_syn_I:
        :param i_offset:
        :param g_leak:
        :param e_rev_E:
        :param gbar_Na:
        :param e_rev_leak:
        :param e_rev_I:
        :param e_rev_K:
        :param v_offset:
        :param v:
        :param gsyn_exc:
        :param gsyn_inh:
        """
        raise SpynnakerException(
            "This neuron model is currently not supported by the tool chain")



# Copyright (c) 2016 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations
import math
from typing import (
    cast, Dict, Iterable, List, Sequence, Tuple, Union, TYPE_CHECKING)

import numpy
from numpy.typing import NDArray
from typing_extensions import TypeAlias

from spinn_utilities.abstract_base import AbstractBase, abstractmethod
from spinn_utilities.overrides import overrides

from pacman.model.graphs.application import ApplicationVertex
from pacman.model.graphs.common import Slice

from spinn_front_end_common.interface.ds import (
    DataType, DataSpecificationBase)
from spinn_front_end_common.utilities.constants import (
    MICRO_TO_MILLISECOND_CONVERSION, MICRO_TO_SECOND_CONVERSION,
    BYTES_PER_WORD, BYTES_PER_SHORT)

from spynnaker.pyNN.data import SpynnakerDataView
from spynnaker.pyNN.exceptions import SynapticConfigurationException
from spynnaker.pyNN.models.common import PopulationApplicationVertex

from .abstract_synapse_dynamics_structural import (
    AbstractSynapseDynamicsStructural)

if TYPE_CHECKING:
    from spynnaker.pyNN.models.projection import Projection
    from spynnaker.pyNN.models.neural_projections import (
        SynapseInformation)
    from spynnaker.pyNN.models.neuron import PopulationVertex
    from spynnaker.pyNN.models.neuron.synaptic_matrices import SynapticMatrices
    from spynnaker.pyNN.models.neuron.synapse_dynamics.types import (
        ConnectionsArray)
    from spynnaker.pyNN.models.neural_projections import (
        ProjectionApplicationEdge)

    _PopIndexType: TypeAlias = Dict[
        Tuple[PopulationApplicationVertex, SynapseInformation], int]
    _SubpopIndexType: TypeAlias = Dict[
        Tuple[PopulationApplicationVertex, SynapseInformation, int], int]

    #: :meta private:
    ConnectionsInfo: TypeAlias = Dict[
        Tuple[PopulationVertex, int],
        List[Tuple[ConnectionsArray, ProjectionApplicationEdge,
                   SynapseInformation]]]

#: Default value for frequency of rewiring
DEFAULT_F_REW = 10 ** 4.0
#: Default value for initial weight on connection formation
DEFAULT_INITIAL_WEIGHT = 0.0
#: Default value for initial delay on connection formation
DEFAULT_INITIAL_DELAY = 1.0
#: Default value for maximum fan-in per target layer neuron
DEFAULT_S_MAX = 32


class SynapseDynamicsStructuralCommon(
        AbstractSynapseDynamicsStructural, metaclass=AbstractBase):
    """
    Common code for structural synapse dynamics.
    """

    # 8 32-bit numbers (fast; p_rew; s_max; app_no_atoms; machine_no_atoms;
    # low_atom; high_atom; with_replacement) + 2 4-word RNG seeds (shared_seed;
    # local_seed) + 1 32-bit number (no_pre_pops)
    _REWIRING_DATA_SIZE = (
        (8 * BYTES_PER_WORD) + (2 * 4 * BYTES_PER_WORD) + BYTES_PER_WORD)

    # Size excluding key_atom_info (as variable length)
    # 4 16-bit numbers (no_pre_vertices; sp_control; delay_lo; delay_hi)
    # + 3 32-bit numbers (weight; connection_type; total_no_atoms)
    _PRE_POP_INFO_BASE_SIZE = (4 * BYTES_PER_SHORT) + (3 * BYTES_PER_WORD)

    # 6 32-bit numbers (key; mask; n_atoms; n_colour_bits; lo_atom;
    # m_pop_index)
    _KEY_ATOM_INFO_SIZE = (6 * BYTES_PER_WORD)

    # 1 16-bit number (neuron_index)
    # + 2 8-bit numbers (sub_pop_index; pop_index)
    _POST_TO_PRE_ENTRY_SIZE = BYTES_PER_SHORT + (2 * 1)

    PAIR_ERROR = (
        "Only one Projection between each pair of Populations can use "
        "structural plasticity")

    __slots__ = ()

    def get_parameter_names(self) -> Iterable[str]:
        """
        :returns: The names of the parameters of the model.
        """
        yield from [
            'initial_weight', 'initial_delay', 'f_rew', 'p_rew', 's_max',
            'with_replacement']
        yield from self.partner_selection.get_parameter_names()
        yield from self.formation.get_parameter_names()
        yield from self.elimination.get_parameter_names()

    @property
    def p_rew(self) -> float:
        """
        The period of rewiring.
        """
        return 1. / self.f_rew

    @overrides(AbstractSynapseDynamicsStructural.write_structural_parameters)
    def write_structural_parameters(
            self, spec: DataSpecificationBase, region: int,
            weight_scales: NDArray[numpy.floating],
            app_vertex: PopulationVertex, vertex_slice: Slice,
            synaptic_matrices: SynapticMatrices) -> None:
        spec.comment("Writing structural plasticity parameters")
        spec.switch_write_focus(region)

        # Get relevant edges
        structural_projections = self.__get_structural_projections(
            app_vertex.incoming_projections)

        # Write the common part of the rewiring data
        self.__write_common_rewiring_data(
            spec, app_vertex, vertex_slice, len(structural_projections))

        # Write the pre-population info
        pop_index, subpop_index, lo_atom_index = \
            self.__write_prepopulation_info(
                spec, app_vertex, structural_projections,
                weight_scales, synaptic_matrices)

        # Write the post-to-pre table
        self.__write_post_to_pre_table(
            spec, pop_index, subpop_index, lo_atom_index, app_vertex,
            vertex_slice)

        # Write the component parameters
        # pylint: disable=protected-access
        spec.comment("Writing partner selection parameters")
        self.partner_selection.write_parameters(spec)
        for proj in structural_projections:
            spec.comment(f"Writing formation parameters for {proj.label}")
            dynamics = cast(AbstractSynapseDynamicsStructural,
                            proj._synapse_information.synapse_dynamics)
            dynamics.formation.write_parameters(spec)
        for proj in structural_projections:
            spec.comment(f"Writing elimination parameters for {proj.label}")
            dynamics = cast(AbstractSynapseDynamicsStructural,
                            proj._synapse_information.synapse_dynamics)
            dynamics.elimination.write_parameters(
                spec, weight_scales[proj._synapse_information.synapse_type])

    def __get_structural_projections(
            self, incoming_projections: Iterable[Projection]
            ) -> List[Projection]:
        structural_projections = list()
        seen_app_edges = set()
        for proj in incoming_projections:
            # pylint: disable=protected-access
            app_edge = proj._projection_edge
            for synapse_info in app_edge.synapse_information:
                if isinstance(synapse_info.synapse_dynamics,
                              AbstractSynapseDynamicsStructural):
                    if app_edge in seen_app_edges:
                        raise SynapticConfigurationException(self.PAIR_ERROR)
                    seen_app_edges.add(app_edge)
                    structural_projections.append(proj)
        return structural_projections

    def __write_common_rewiring_data(
            self, spec: DataSpecificationBase,
            app_vertex: PopulationVertex, vertex_slice: Slice,
            n_pre_pops: int) -> None:
        """
        Write the non-sub-population synapse parameters to the spec.

        :param spec: the data spec
        :param app_vertex: The application vertex being generated
        :param vertex_slice: The slice of the target vertex to generate for
        :param n_pre_pops: the number of pre-populations
        """
        time_step_us = SpynnakerDataView.get_simulation_time_step_us()
        spec.comment("Writing common rewiring data")
        if (self.p_rew * MICRO_TO_MILLISECOND_CONVERSION <
                time_step_us / MICRO_TO_MILLISECOND_CONVERSION):
            # Fast rewiring
            spec.write_value(data=1)
            spec.write_value(data=int(
                time_step_us / (
                    self.p_rew * MICRO_TO_SECOND_CONVERSION)))
        else:
            # Slow rewiring
            spec.write_value(data=0)
            spec.write_value(data=int((
                self.p_rew * MICRO_TO_SECOND_CONVERSION) /
                time_step_us))
        # write s_max
        spec.write_value(data=int(self.s_max))
        # write total number of atoms in the application vertex
        spec.write_value(data=app_vertex.n_atoms)
        # write local low, high and number of atoms
        spec.write_value(data=vertex_slice.n_atoms)
        spec.write_value(data=vertex_slice.lo_atom)
        spec.write_value(data=vertex_slice.hi_atom)
        # write with_replacement
        spec.write_value(data=self.with_replacement)

        # write app level seeds
        spec.write_array(self._get_seeds(app_vertex))

        # write local seed (4 words), generated randomly!
        # Note that in case of a reset, these need a key to ensure subsequent
        # runs match the first run
        spec.write_array(self._get_seeds(vertex_slice))

        # write the number of pre-populations
        spec.write_value(data=n_pre_pops)

    def __write_prepopulation_info(
            self, spec: DataSpecificationBase,
            app_vertex: ApplicationVertex,
            structural_projections: Iterable[Projection],
            weight_scales: NDArray[numpy.floating],
            synaptic_matrices: SynapticMatrices) -> Tuple[
                _PopIndexType, _SubpopIndexType, _SubpopIndexType]:
        """
        :param spec:
        :param app_vertex:
            the vertex for which data specs are being prepared
        :param structural_projections: Projections that are structural
        :param weight_scales:
        :param synaptic_matrices:
        """
        spec.comment("Writing pre-population info")
        pop_index: _PopIndexType = dict()
        routing_info = SpynnakerDataView.get_routing_infos()
        subpop_index: _SubpopIndexType = dict()
        lo_atom_index: _SubpopIndexType = dict()
        index = 0
        for proj in structural_projections:
            spec.comment(f"Writing pre-population info for {proj.label}")
            # pylint: disable=protected-access
            app_edge = proj._projection_edge
            synapse_info = proj._synapse_information
            pop_index[app_edge.pre_vertex, synapse_info] = index
            index += 1
            dynamics = cast(AbstractSynapseDynamicsStructural,
                            synapse_info.synapse_dynamics)

            # Number of incoming vertices
            out_verts = app_edge.pre_vertex.splitter.get_out_going_vertices(
                synapse_info.partition_id)
            spec.write_value(len(out_verts), data_type=DataType.UINT16)

            # Controls - currently just if this is a self connection or not
            self_connected = app_vertex == app_edge.pre_vertex
            spec.write_value(int(self_connected), data_type=DataType.UINT16)
            # Delay
            delay_scale = SpynnakerDataView.get_simulation_time_step_per_ms()
            if isinstance(dynamics.initial_delay, tuple):
                spec.write_value(int(dynamics.initial_delay[0] * delay_scale),
                                 data_type=DataType.UINT16)
                spec.write_value(int(dynamics.initial_delay[1] * delay_scale),
                                 data_type=DataType.UINT16)
            else:
                scaled_delay = dynamics.initial_delay * delay_scale
                spec.write_value(scaled_delay, data_type=DataType.UINT16)
                spec.write_value(scaled_delay, data_type=DataType.UINT16)

            # Weight
            spec.write_value(round(dynamics.initial_weight *
                                   weight_scales[synapse_info.synapse_type]))
            # Connection type
            spec.write_value(synapse_info.synapse_type)
            # Total number of atoms in pre-vertex
            spec.write_value(app_edge.pre_vertex.n_atoms)
            # Machine edge information
            for sub, m_vertex in enumerate(out_verts):
                r_info = routing_info.get_info_from(
                    m_vertex, synapse_info.partition_id)
                vertex_slice = m_vertex.vertex_slice
                spec.write_value(r_info.key)
                spec.write_value(r_info.mask)
                out_app = m_vertex.app_vertex
                assert isinstance(out_app, PopulationApplicationVertex)
                spec.write_value(out_app.n_colour_bits)
                spec.write_value(vertex_slice.n_atoms)
                spec.write_value(vertex_slice.lo_atom)
                spec.write_value(synaptic_matrices.get_index(
                    app_edge, synapse_info))
                lo = vertex_slice.lo_atom
                for i in range(lo, vertex_slice.hi_atom + 1):
                    subpop_index[app_edge.pre_vertex, synapse_info, i] = sub
                    lo_atom_index[app_edge.pre_vertex, synapse_info, i] = lo
        return pop_index, subpop_index, lo_atom_index

    def __write_post_to_pre_table(
            self, spec: DataSpecificationBase, pop_index: _PopIndexType,
            subpop_index: _SubpopIndexType, lo_atom_index: _SubpopIndexType,
            app_vertex: PopulationVertex, vertex_slice: Slice) -> None:
        """
        Post to pre table is basically the transpose of the synaptic matrix.

        :param spec:
        :param pop_index:
        :param app_vertex: the vertex for which data specs are being prepared
        :param vertex_slice: The target slice
        """
        # Get connections for this post slice
        slice_conns = self.connections[app_vertex, vertex_slice.lo_atom]
        # Make a single large array of connections
        connections = numpy.concatenate(
            [conn for (conn, _, _) in slice_conns])
        # Make a single large array of population index
        conn_lens = [len(conn) for (conn, _, _) in slice_conns]
        for (_, a_edge, s_info) in slice_conns:
            if (a_edge.pre_vertex, s_info) not in pop_index:
                print("Help!")
        pop_indices = numpy.repeat(
            [pop_index[a_edge.pre_vertex, s_info]
             for (_, a_edge, s_info) in slice_conns], conn_lens)
        # Make a single large array of sub-population index
        subpop_indices = numpy.array([
            subpop_index[a_edge.pre_vertex, s_info, c["source"]]
            for (conns, a_edge, s_info) in slice_conns for c in conns])
        # Get the low atom for each source and subtract
        lo_atoms = numpy.array([
            lo_atom_index[a_edge.pre_vertex, s_info, c["source"]]
            for (conns, a_edge, s_info) in slice_conns for c in conns])
        connections["source"] = connections["source"] - lo_atoms

        # Make an array of all data required
        conn_data = numpy.dstack(
            (pop_indices, subpop_indices, connections["source"]))[0]

        # Break data into rows based on target and strip target out
        rows = [conn_data[connections["target"] == i]
                for i in range(0, vertex_slice.n_atoms)]

        if any(len(row) > self.s_max for row in rows):
            raise ValueError(
                "Too many initial connections per incoming neuron")

        # Make each row the required length through padding with 0xFFFF
        padded_rows = [numpy.pad(row, [(self.s_max - len(row), 0), (0, 0)],
                                 "constant", constant_values=0xFFFF)
                       for row in rows]

        # Finally make the table and write it out
        post_to_pre = numpy.rec.fromarrays(
            numpy.concatenate(padded_rows).T, formats="u1, u1, u2").view("u4")
        if len(post_to_pre) != vertex_slice.n_atoms * self.s_max:
            raise ValueError(
                f"Wrong size of pre-to-pop tables: {len(post_to_pre)} "
                f"Found, {vertex_slice.n_atoms * self.s_max} Expected")
        spec.comment(
            "Writing post-to-pre table of "
            f"{vertex_slice.n_atoms * self.s_max} words")
        spec.write_array(post_to_pre)

    @overrides(AbstractSynapseDynamicsStructural.
               get_structural_parameters_sdram_usage_in_bytes)
    def get_structural_parameters_sdram_usage_in_bytes(
            self, incoming_projections: Iterable[Projection],
            n_neurons: int) -> int:
        # Work out how many sub-edges we will end up with, as this is used
        # for key_atom_info
        param_sizes = (
            self.partner_selection.get_parameters_sdram_usage_in_bytes())
        n_sub_edges = 0
        structural_projections = self.__get_structural_projections(
            incoming_projections)
        for proj in structural_projections:
            # pylint: disable=protected-access
            dynamics = cast(AbstractSynapseDynamicsStructural,
                            proj._synapse_information.synapse_dynamics)
            app_edge = proj._projection_edge
            n_sub_edges += len(
                app_edge.pre_vertex.splitter.get_out_going_slices())
            param_sizes += dynamics.formation\
                .get_parameters_sdram_usage_in_bytes()
            param_sizes += dynamics.elimination\
                .get_parameters_sdram_usage_in_bytes()

        return int(
            self._REWIRING_DATA_SIZE +
            (self._PRE_POP_INFO_BASE_SIZE * len(structural_projections)) +
            (self._KEY_ATOM_INFO_SIZE * n_sub_edges) +
            (self._POST_TO_PRE_ENTRY_SIZE * n_neurons * self.s_max) +
            param_sizes)

    def get_vertex_executable_suffix(self) -> str:
        """
        :returns: executable suffix based on details
        """
        name = "_structural"
        name += self.partner_selection.vertex_executable_suffix
        name += self.formation.vertex_executable_suffix
        name += self.elimination.vertex_executable_suffix
        return name

    def is_same_as(
            self, synapse_dynamics: AbstractSynapseDynamicsStructural) -> bool:
        """
        :param synapse_dynamics:
        :returns: True if the two have the same data and types
        """
        # Note noqa:E721  because exact type comparison is required here
        return (
            self.s_max == synapse_dynamics.s_max and
            self.f_rew == synapse_dynamics.f_rew and
            self.initial_weight == synapse_dynamics.initial_weight and
            self.initial_delay == synapse_dynamics.initial_delay and
            # pylint: disable=unidiomatic-typecheck
            (type(self.partner_selection) ==  # noqa: E721
             type(synapse_dynamics.partner_selection)) and
            (type(self.formation) ==
             type(synapse_dynamics.formation)) and
            (type(self.elimination) ==
             type(synapse_dynamics.elimination)))

    @property
    @abstractmethod
    def connections(self) -> ConnectionsInfo:
        """
        Initial connectivity as defined via connector.
        """
        raise NotImplementedError

    @abstractmethod
    def _get_seeds(
            self, app_vertex: Union[None, ApplicationVertex, Slice] = None
            ) -> Sequence[int]:
        """
        Generate a seed for the RNG on chip that is the same for all
        of the cores that have perform structural updates.

        It should be different between application vertices
        but the same for the same app_vertex.
        It should be different every time called with `None`.

        :return: list of random seed (4 words), generated randomly
        """
        raise NotImplementedError

    @overrides(AbstractSynapseDynamicsStructural.check_initial_delay)
    def check_initial_delay(self, max_delay_ms: float) -> None:
        """
        Check that delays can be done without delay extensions.

        :param max_delay_ms: The maximum delay supported, in milliseconds
        :raises ValueError: if the delay is out of range
        """
        if isinstance(self.initial_delay, tuple):
            init_del = self.initial_delay
            if init_del[0] > max_delay_ms or init_del[1] > max_delay_ms:
                raise ValueError(
                    f"The initial delay {self.initial_delay} has one or more "
                    f"values that are bigger than {max_delay_ms}.")
        elif self.initial_delay > max_delay_ms:
            raise ValueError(
                f"The initial delay {self.initial_delay} "
                f"is bigger than {max_delay_ms}.")

    def get_max_rewires_per_ts(self) -> int:
        max_rewires_per_ts = 1
        if (self.p_rew * MICRO_TO_MILLISECOND_CONVERSION <
                SpynnakerDataView.get_simulation_time_step_ms()):
            # fast rewiring, so need to set max_rewires_per_ts
            max_rewires_per_ts = int(
                SpynnakerDataView.get_simulation_time_step_us() / (
                        self.p_rew * MICRO_TO_SECOND_CONVERSION))

        return max_rewires_per_ts

    @property
    def reduction_synapses_per_second(self) -> int:
        """
        Approximate fewer number of synapses that can be processed per second
        as a result of rewiring cycles used.
        """
        # Guess that each rewiring attempt takes the same as 17 synapses,
        # based on ~250 cycles per rewiring attempt, and 15 cycles per synapse
        return math.ceil(17 * self.f_rew)



# Copyright (c) 2017 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import annotations
import logging
import os
from typing import (
    Any, Callable, Dict, Iterable, Iterator, List, Optional,
    overload, Sequence, Tuple, TYPE_CHECKING, Union)

import numpy
from numpy import bool_, integer
from numpy.typing import NDArray

from pyNN import descriptions
from pyNN.random import NumpyRNG
from pyNN.space import BaseStructure
import neo  # type: ignore[import]

from spinn_utilities.log import FormatAdapter
from spinn_utilities.logger_utils import warn_once
from spinn_utilities.overrides import overrides
from spinn_utilities.ranged.abstract_sized import AbstractSized

from spynnaker.pyNN.utilities.neo_buffer_database import NeoBufferDatabase
from spynnaker.pyNN.types import Selector
from spynnaker.pyNN.utilities.utility_calls import get_neo_io

from .population_base import PopulationBase

if TYPE_CHECKING:
    from .population import Population
    from spynnaker.pyNN.models.current_sources import AbstractCurrentSource
    from spynnaker.pyNN.models.abstract_pynn_model import AbstractPyNNModel
    from spynnaker.pyNN.models.common.population_application_vertex import (
        PopulationApplicationVertex)
    from spynnaker.pyNN.models.common import ParameterHolder
    from spynnaker.pyNN.models.common.types import Names, Values
    from spynnaker.pyNN.models.recorder import Recorder
    from spynnaker.pyNN.types import IoDest

logger = FormatAdapter(logging.getLogger(__name__))


class PopulationView(PopulationBase):
    """
    A view of a subset of neurons within a
    :py:class:`~spynnaker.pyNN.models.populations.Population`.

    In most ways, Populations and PopulationViews have the same behaviour,
    i.e., they can be recorded, connected with Projections, etc.
    It should be noted that any changes to neurons in a PopulationView
    will be reflected in the parent Population and *vice versa.*

    It is possible to have views of views.

    .. note::
        Selector to Id is actually handled by :py:class:`~.AbstractSized`.
    """
    __slots__ = (
        "__annotations",
        "__indexes",
        "__label",
        "__mask",
        "__parent",
        "__population",
        "__vertex",
        "__recorder")

    __realslots__ = frozenset("_PopulationView" + item for item in __slots__)

    def __init__(
            self, parent: Union[Population, 'PopulationView'],
            selector: Selector, label: Optional[str] = None):
        """
        :param parent: the population or view to make the view from
        :param selector: a slice or numpy mask array.
            The mask array should either be a Boolean array (ideally) of the
            same size as the parent,
            or an integer array containing cell indices,
            i.e. if `p.size == 5` then:

            ::

                PopulationView(p, array([False, False, True, False, True]))
                PopulationView(p, array([2, 4]))
                PopulationView(p, slice(2, 5, 2))

            will all create the same view.
        :param label: A label for the view
        """
        self.__parent = parent
        sized = AbstractSized(parent.size)
        ids = sized.selector_to_ids(selector, warn=True)

        if isinstance(parent, PopulationView):
            self.__population = parent.grandparent
            self.__indexes: Sequence[int] = parent.index_in_grandparent(ids)
        else:
            self.__population = parent
            self.__indexes = ids
        self.__mask = selector
        if label is None:
            label = f"{parent.label}:{selector}"
        self.__label = label
        self.__annotations: Dict[str, Any] = dict()

        # Get these two objects to make access easier
        # pylint: disable=protected-access
        self.__vertex = self.__population._vertex
        # pylint: disable=protected-access
        self.__recorder = self.__population._recorder

    def __getattr__(self, name: str) -> ParameterHolder:
        return self.__vertex.get_parameter_values(name, self.__indexes)

    def __setattr__(self, name: str, value: Values) -> None:
        if name in self.__realslots__:
            object.__setattr__(self, name, value)
            return
        return self.__vertex.set_parameter_values(name, value, self.__indexes)

    @property
    def size(self) -> int:
        """
        The total number of neurons in the Population View.
        """
        return len(self.__indexes)

    @property
    def label(self) -> str:
        """
        A label for the Population View.
        """
        return self.__label

    @property
    def celltype(self) -> AbstractPyNNModel:
        """
        The type of neurons making up the underlying Population.
        """
        return self.__parent.celltype

    @property
    def initial_values(self) -> ParameterHolder:
        """
        A dict containing the initial values of the state variables.
        """
        return self.__vertex.get_initial_state_values(
            self.__vertex.get_state_variables(), self.__indexes)

    @property
    def current_values(self) -> ParameterHolder:
        """
        A dict containing the current values of the state variables.
        """
        warn_once(
            logger, "current_values is non-standard PyNN and therefore "
            "will not be portable to other simulators.")
        return self.__vertex.get_current_state_values(
            self.__vertex.get_state_variables(), self.__indexes)

    @property
    def parent(self) -> Union[Population, 'PopulationView']:
        """
        A reference to the parent Population (that this is a view of).
        """
        return self.__parent

    @property
    def mask(self) -> Selector:
        """
        The selector mask that was used to create this view.
        """
        return self.__mask

    @property
    @overrides(PopulationBase.all_cells)
    def all_cells(self) -> Sequence['IDMixin']:
        """
        An array containing the cell IDs of all neurons in the
        Population (all MPI nodes).
        """
        return [IDMixin(self.__population, idx) for idx in self.__indexes]

    @property
    def _indexes(self) -> Tuple[int, ...]:
        return tuple(self.__indexes)

    @overload
    def __getitem__(self, index: int) -> 'IDMixin':
        ...

    @overload
    def __getitem__(self, index: Union[
            None, slice, List[int], List[bool], NDArray[bool_],
            NDArray[integer]]) -> 'PopulationView':
        ...

    def __getitem__(self, index: Union[
            None, int, slice, List[int], List[bool], NDArray[bool_],
            NDArray[integer]]) -> 'PopulationView':
        """
        Return either a single cell (ID object) from the Population,
        if index is an integer, or a subset of the cells
        (PopulationView object), if index is a slice or array.

        .. note::
            ``__getitem__`` is called when using[] access, e.g. if
            ``p = Population(...)`` then
            ``p[2]`` is equivalent to ``p.__getitem__(2)``, and
            ``p[3:6]`` is equivalent to ``p.__getitem__(slice(3, 6))``
        """
        if isinstance(index, int):
            return IDMixin(self.__population, index)
        return PopulationView(self, index, label=self.label + "_" + str(index))

    def __iter__(self) -> Iterator['IDMixin']:
        """
        Iterator over cell IDs (on the local node).
        """
        for idx in self.__indexes:
            yield IDMixin(self.__population, idx)

    def __len__(self) -> int:
        """
        Return the total number of cells in the population (all nodes).
        """
        return len(self.__indexes)

    @property
    @overrides(PopulationBase.positions)
    def positions(self) -> NDArray[numpy.floating]:
        raise NotImplementedError("Not implemented for views")

    @property
    @overrides(PopulationBase.position_generator)
    def position_generator(self) -> Callable[[int], NDArray[numpy.floating]]:
        raise NotImplementedError("Not implemented for views")

    def all(self) -> Iterator[IDMixin]:
        """
        :returns: An iterator over cell IDs (on all MPI nodes).
        """
        for idx in self.__indexes:
            yield IDMixin(self.__population, idx)

    def can_record(self, variable: str) -> bool:
        """
        Determine whether variable can be recorded from this population.

        :returns: True if this variable can be recorded,
            False otherwise including if the variable is invalid.
        """
        return variable in self.__vertex.get_recordable_variables()

    @property
    def conductance_based(self) -> bool:
        """
        Whether the post-synaptic response is modelled as a
        change in conductance or a change in current.
        """
        return self.__vertex.conductance_based

    @overrides(PopulationBase.inject)
    def inject(self, current_source: AbstractCurrentSource) -> None:
        """
        Injects the specified current_source into this PopulationView.

        :param current_source:
            the current source to be injected
        """
        self.__vertex.inject(current_source, self.__indexes)

    def describe(self, template: str = 'populationview_default.txt',
                 engine: str = 'default') -> str:
        """
        Returns a human-readable description of the population view.

        The output may be customised by specifying a different template
        together with an associated template engine (see pyNN.descriptions).

        If template is ``None``, then a dictionary containing the template
        context will be returned.

        :param template: Template filename
        :param engine: Template substitution engine
        :returns: A human-readable description of the population view.
        """
        context = {"label": self.label,
                   "parent": self.parent.label,
                   "mask": self.mask,
                   "size": self.size}
        context.update(self.__annotations)
        return descriptions.render(engine, template, context)

    def find_units(self, variable: str) -> str:
        """
        Get the units of a variable

        .. warning::
            No PyNN description of this method.

        :param variable: The name of the variable
        :return: The units of the variable
        """
        return self.__vertex.get_units(variable)

    def get(self, parameter_names: Names,
            gather: bool = False, simplify: bool = True) -> ParameterHolder:
        """
        Get the values of the given parameters for all cells in the population.

        Values will be expressed in the standard PyNN units (i.e. millivolts,
        nanoamps, milliseconds, microsiemens, nanofarads, event per second).

        .. note::
            SpiNNaker always gathers.

        :param parameter_names:
        :param gather: Ignored. Purely for PyNN compatibility
        :param simplify: Ignored.  Purely for PyNN compatibility
        :returns: The values of the given parameters for all cell
        """
        if not gather:
            logger.warning("SpiNNaker only supports gather=True. We will run "
                           "as if gather was set to True.")
        if simplify is not True:
            logger.warning("The simplify value is ignored if not set to true")

        return self.__vertex.get_parameter_values(
            parameter_names, self.__indexes)

    @overrides(PopulationBase.get_data)
    def get_data(
            self, variables: Names = 'all',
            gather: bool = True, clear: bool = False, *,
            annotations: Optional[Dict[str, Any]] = None) -> neo.Block:
        """
        Return a Neo Block containing the data(spikes, state variables)
        recorded from the Population.

        :param variables: Either a single variable name or a list of variable
            names. Variables must have been previously recorded, otherwise an
            Exception will be raised.
        :param gather: Ignored. Purely for PyNN compatibility

            .. note::
                SpiNNaker always gathers.

        :param clear:
            If True, recorded data will be deleted from the Population.
        :param annotations: annotations to put on the neo block
        :raises \
            ~spinn_front_end_common.utilities.exceptions.ConfigurationException:
            If the variable or variables have not been previously set to
            record.
        """
        if not gather:
            logger.warning("SpiNNaker only supports gather=True. We will run "
                           "as if gather was set to True.")
        if annotations is not None:
            warn_once(
                logger, "Annotations parameter is not standard PyNN so may "
                "not be supported by all platforms.")

        return self.__recorder.extract_neo_block(
            variables, self.__indexes, clear, annotations)

    def spinnaker_get_data(
            self, variable: str, as_matrix: bool = False) -> NDArray:
        """
        SsPyNNaker specific method for getting data as a numpy array,
        instead of the Neo-based object

        :param variable: a single variable name
        :param as_matrix: If set True the data is returned as a 2d matrix
        :return: array of the data
        """
        return self.__population.spinnaker_get_data(
            variable, as_matrix, self.__indexes)

    @overrides(PopulationBase.get_spike_counts)
    def get_spike_counts(self, gather: bool = True) -> Dict[int, int]:
        self._check_params(gather)
        with NeoBufferDatabase() as db:
            return db.get_spike_counts(
                self.__recorder.recording_label, self.__indexes)

    @property
    def grandparent(self) -> Population:
        """
        The parent Population at the root of the tree (since the
        immediate parent may itself be a PopulationView).

        The name "grandparent" is of course a little misleading, as it could
        be just the parent, or the great, great, great, ..., grandparent.
        """
        return self.__population

    @overload
    def id_to_index(self, id: int) -> int:  # pylint: disable=redefined-builtin
        ...

    @overload
    def id_to_index(
            self, id: Iterable[int]  # pylint: disable=redefined-builtin
            ) -> List[int]:
        ...

    def id_to_index(
            self, id: Union[int, Iterable[int]]) -> \
            Union[int, List[int]]:  # pylint: disable=redefined-builtin
        """
        Given the ID(s) of cell(s) in the PopulationView, return its /
        their index / indices(order in the PopulationView).

        assert pv.id_to_index(pv[3]) == 3

        :param id:
        :returns: Index in this View
        """
        if isinstance(id, int):
            return self.__indexes.index(id)
        return [self.__indexes.index(idx) for idx in id]

    def index_in_grandparent(self, indices: Iterable[int]) -> List[int]:
        """
        Given an array of indices, return the indices in the parent
        population at the root of the tree.

        :param indices:
        :returns: Indices in the parent
        """
        return [self.__indexes[index] for index in indices]

    def initialize(self, **initial_values: Values) -> None:
        """
        Set initial values of state variables, e.g. the membrane potential.
        Values passed to ``initialize()`` may be:

        * single numeric values (all neurons set to the same value), or
        * :py:class:`~spynnaker.pyNN.RandomDistribution` objects, or
        * lists / arrays of numbers of the same size as the population
          mapping functions, where a mapping function accepts a single
          argument (the cell index) and returns a single number.

        Values should be expressed in the standard PyNN units (i.e.
        millivolts, nanoamps, milliseconds, microsiemens, nanofarads,
        events per second).

        Examples::

            p.initialize(v=-70.0)
            p.initialize(v=rand_distr, gsyn_exc=0.0)
            p.initialize(v=lambda i: -65 + i / 10.0)
        """
        for variable, value in initial_values.items():
            self.__vertex.set_initial_state_values(
                variable, value, self.__indexes)

    def set_state(self, **initial_values: Values) -> None:
        """
        Set current values of state variables, e.g. the membrane potential.
        Values passed to ``set_state()`` may be:

        * single numeric values (all neurons set to the same value), or
        * :py:class:`~spynnaker.pyNN.RandomDistribution` objects, or
        * lists / arrays of numbers of the same size as the population
          mapping functions, where a mapping function accepts a single
          argument (the cell index) and returns a single number.

        Values should be expressed in the standard PyNN units (i.e.
        millivolts, nanoamps, milliseconds, microsiemens, nanofarads,
        events per second).

        Examples::

            p.set_state(v=-70.0)
            p.set_state(v=rand_distr, gsyn_exc=0.0)
            p.set_state(v=lambda i: -65 + i / 10.0)
        """
        warn_once(
            logger, "set_state is non-standard PyNN and therefore "
            "will not be portable to other simulators.")
        for variable, value in initial_values.items():
            self.__vertex.set_current_state_values(
                variable, value, self.__indexes)

    @overrides(PopulationBase.record)
    def record(self, variables: Names, to_file: IoDest = None,
               sampling_interval: Optional[int] = None) -> None:
        self.__recorder.record(
            variables, to_file, sampling_interval, self.__indexes)

    @property
    @overrides(PopulationBase.structure)
    def structure(self) -> Optional[BaseStructure]:
        raise NotImplementedError("Not implemented for views")

    def sample(
            self, n: int, rng: Optional[NumpyRNG] = None) -> 'PopulationView':
        """
        Randomly sample `n` cells from the Population view, and return a
        new PopulationView object.

        :param n: The number of cells to select
        :param rng: Random number generator
        :returns: A PopulationView over n random cells.
        """
        if not rng:
            rng = NumpyRNG()
        indices = rng.permutation(
            numpy.arange(len(self), dtype=numpy.int32))[0:n]
        return PopulationView(
            self, indices,
            label=f"Random sample size {n} from {self.label}")

    def set(self, **parameters: Values) -> None:
        """
        Set one or more parameters for every cell in the population.
        Values passed to `set()` may be:

        * single values,
        * :py:class:`~spynnaker.pyNN.RandomDistribution` objects, or
        * lists / arrays of values of the same size as the population
          mapping functions, where a mapping function accepts a single
          argument (the cell index) and returns a single value.

        Here, a "single value" may be either a single number or a list /
        array of numbers (e.g. for spike times).

        Values should be expressed in the standard PyNN units (i.e.
        millivolts, nanoamps, milliseconds, microsiemens, nanofarads,
        event per second).

        Examples::

            p.set(tau_m=20.0, v_rest=-65).
            p.set(spike_times=[0.3, 0.7, 0.9, 1.4])
            p.set(cm=rand_distr, tau_m=lambda i: 10 + i / 10.0)
        """
        for (parameter, value) in parameters.items():
            self.__vertex.set_parameter_values(
                parameter, value, self.__indexes)

    @overrides(PopulationBase.write_data)
    def write_data(self, io: Union[str, neo.baseio.BaseIO],
                   variables: Names = 'all',
                   gather: bool = True, clear: bool = False,
                   annotations: Optional[Dict[str, Any]] = None) -> None:
        if not gather:
            logger.warning("SpiNNaker only supports gather=True. We will run "
                           "as if gather was set to True.")
        if isinstance(io, str):
            extension = os.path.splitext(io)[1][1:]
            if extension == "csv":
                self.__recorder.write_data(
                    io, variables, view_indexes=self.__indexes,
                    annotations=annotations)
                return
            io = get_neo_io(io)

        data = self.__recorder.extract_neo_block(
            variables, self.__indexes, clear, annotations)

        # write the neo block to the file
        io.write(data)

    @property
    @overrides(PopulationBase._vertex)
    def _vertex(self) -> PopulationApplicationVertex:
        return self.__vertex

    @property
    @overrides(PopulationBase._recorder)
    def _recorder(self) -> Recorder:
        return self.__recorder

    @property
    @overrides(PopulationBase._view_range)
    def _view_range(self) -> Tuple[int, int]:
        indices = self.__indexes
        return indices[0], indices[-1]

    @property
    def _is_contiguous(self) -> bool:
        cont = range(self.__indexes[0], self.__indexes[-1] + 1)
        # self.__indexes is likely a range too so test direct first
        if self.__indexes == cont:
            return True
        return tuple(self.__indexes) == tuple(cont)

    def __eq__(self, other: Any) -> bool:
        if not isinstance(other, PopulationView):
            return False
        return (self.__vertex == other._vertex and
                self._indexes == other._indexes)

    def __str__(self) -> str:
        return str(self.__vertex) + str(self.__indexes)

    def __repr__(self) -> str:
        return repr(self.__vertex) + str(self.__indexes)


class IDMixin(PopulationView):
    """
    Implementation of PyNN IDMixin.
    """
    __slots__ = ()

    def get_parameters(self) -> ParameterHolder:
        """
        :returns: A dict of all cell parameters.
        """
        return self._vertex.get_parameter_values(
            self._vertex.get_parameters(), self.id)

    # NON-PYNN API CALLS
    @property
    def id(self) -> int:
        """
        The lowest index in this view
        """
        return self._indexes[0]

    def __getattr__(self, name: str) -> ParameterHolder:
        if name == "_vertex":
            raise KeyError("Shouldn't come through here!")
        return self._vertex.get_parameter_values(name, self.id)

    def __setattr__(self, name: str, value: Values) -> None:
        if name in self.__realslots__:
            object.__setattr__(self, name, value)
            return
        return self._vertex.set_parameter_values(name, value, self.id)

    def get_initial_value(self, variable: str) -> ParameterHolder:
        """
        Get the initial value of a state variable of the cell.

        :param variable: The name of the variable
        :returns: The initial value for this variable
        """
        return self._vertex.get_initial_state_values(variable, self.id)

    @property
    def initial_values(self) -> ParameterHolder:
        return self._vertex.get_initial_state_values(
            self._vertex.get_state_variables(), self.id)

    def set_initial_value(self, variable: str, value: Values) -> None:
        """
        Set the initial value of a state variable of the cell.

        :param variable: The name of the variable
        :param value: The value of the variable
        """
        self._vertex.set_initial_state_values(variable, value, self.id)

    def set_parameters(self, **parameters: Values) -> None:
        """
        Set cell parameters, given as a sequence of parameter=value arguments.
        """
        for (name, value) in parameters.items():
            self._vertex.set_parameter_values(name, value, self.id)

    def as_view(self) -> PopulationView:
        """
        :returns: A PopulationView containing just this cell.
        """
        return self

    @property
    def local(self) -> bool:
        """
        Whether this cell is local to the current MPI node.
        """
        # There are no MPI nodes!
        return True

    @property
    @overrides(PopulationBase.positions)
    def positions(self) -> NDArray[numpy.floating]:
        raise NotImplementedError("Not implemented for IDMixin")

    @property
    @overrides(PopulationBase.position_generator)
    def position_generator(self) -> Callable[[int], NDArray[numpy.floating]]:
        raise NotImplementedError("Not implemented for IDMixin")

    @property
    @overrides(PopulationBase.structure)
    def structure(self) -> Optional[BaseStructure]:
        raise NotImplementedError("Not implemented for IDMixin")



# Copyright (c) 2017 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Any, Dict, Optional

from spinn_utilities.overrides import overrides

from spynnaker.pyNN.external_devices_models import AbstractEthernetSensor
from spynnaker.pyNN.external_devices_models.push_bot import (
    AbstractPushBotRetinaDevice)
from spynnaker.pyNN.external_devices_models.push_bot.parameters import (
    PushBotRetinaResolution)
from spynnaker.pyNN.protocols import MunichIoSpiNNakerLinkProtocol
from .push_bot_retina_connection import PushBotRetinaConnection
from .push_bot_translator import PushBotTranslator
from .push_bot_wifi_connection import get_pushbot_wifi_connection


class PushBotEthernetRetinaDevice(
        AbstractPushBotRetinaDevice, AbstractEthernetSensor):
    """
    A PushBot retina over Ethernet
    """
    def __init__(
            self, protocol: MunichIoSpiNNakerLinkProtocol,
            resolution: PushBotRetinaResolution,
            pushbot_ip_address: str, pushbot_port: int = 56000,
            injector_port: Optional[int] = None,
            local_host: Optional[str] = None,
            local_port: Optional[int] = None,
            retina_injector_label: str = "PushBotRetinaInjector"):
        """
        :param protocol:
        :param resolution:
        :param pushbot_ip_address:
        :param pushbot_port:
        :param injector_port:
        :param local_host:
        :param local_port:
        :param retina_injector_label:
        """
        super().__init__(protocol, None)
        pushbot_wifi_connection = get_pushbot_wifi_connection(
            pushbot_ip_address, pushbot_port)
        self.__translator = PushBotTranslator(
            protocol, pushbot_wifi_connection)
        self.__injector_port = injector_port
        self.__retina_injector_label = retina_injector_label

        self.__database_connection = PushBotRetinaConnection(
            self.__retina_injector_label, pushbot_wifi_connection, resolution,
            local_host, local_port)
        self.__n_neurons = resolution.value.n_neurons

    @overrides(AbstractEthernetSensor.get_n_neurons)
    def get_n_neurons(self) -> int:
        return self.__n_neurons

    @overrides(AbstractEthernetSensor.get_injector_parameters)
    def get_injector_parameters(self) -> Dict[str, Any]:
        return {"port": self.__injector_port}

    @overrides(AbstractEthernetSensor.get_injector_label)
    def get_injector_label(self) -> str:
        return self.__retina_injector_label

    @overrides(AbstractEthernetSensor.get_translator)
    def get_translator(self) -> PushBotTranslator:
        return self.__translator

    @overrides(AbstractEthernetSensor.get_database_connection)
    def get_database_connection(self) -> PushBotRetinaConnection:
        return self.__database_connection



# Copyright (c) 2017 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import itertools
import logging
import math
from types import MappingProxyType
from typing import (
    Collection, Dict, Iterable, List, Mapping, Optional, Sequence, Set, Tuple)

import numpy
from numpy import float64, uint8, uint16, uint32
from numpy.typing import NDArray

from spinn_utilities.log import FormatAdapter

from pacman.model.graphs.application import ApplicationVertex
from pacman.model.graphs.machine import MachineVertex
from pacman.model.graphs.common import Slice
from pacman.model.resources import AbstractSDRAM, VariableSDRAM
from pacman.model.placements import Placement

from spinn_front_end_common.interface.ds import (
    DataType, DataSpecificationBase)
from spinn_front_end_common.utilities.exceptions import ConfigurationException
from spinn_front_end_common.utilities.constants import (
    BYTES_PER_WORD, BITS_PER_WORD)

from spynnaker.pyNN.data import SpynnakerDataView
from spynnaker.pyNN.utilities.buffer_data_type import BufferDataType

logger = FormatAdapter(logging.getLogger(__name__))

# The number to add to generate that all neurons are in the given state
_REPEAT_PER_NEURON_RECORDED = 0x7FFFFFFF

# The number to add to generate that all neurons are recorded
_REPEAT_PER_NEURON = 0xFFFFFFFF

# The flag to add to generate that the count is recorded
_RECORDED_FLAG = 0x80000000

# The flag (or lack thereof) to add to generate that the count is not recorded
_NOT_RECORDED_FLAG = 0x00000000


def get_sampling_interval(sampling_rate: float) -> float:
    """
    Return the current sampling interval given a sampling rate.

    :param sampling_rate: The sampling rate in time steps
    :return: Sampling interval in microseconds
    """
    return sampling_rate * SpynnakerDataView.get_simulation_time_step_ms()


class NeuronRecorder(object):
    """
    Methods related to recording of neuron data.
    """
    __slots__ = (
        "__indexes",
        "__n_neurons",
        "__sampling_rates",
        "__data_types",
        "__bitfield_variables",
        "__per_timestep_variables",
        "__per_timestep_datatypes",
        "__per_timestep_recording",
        "__events_per_core_variables",
        "__events_per_core_datatypes",
        "__events_per_core_recording",
        "__events_per_ts",
        "__region_ids",
        "__offset_added")

    _N_BYTES_FOR_TIMESTAMP = BYTES_PER_WORD
    _N_BYTES_PER_RATE = BYTES_PER_WORD
    _N_BYTES_PER_ENUM = BYTES_PER_WORD
    _N_BYTES_PER_GEN_ITEM = BYTES_PER_WORD

    #: size of a index in terms of position into recording array
    _N_BYTES_PER_INDEX = DataType.UINT16.size  # currently uint16

    #: size of the counter for spike recording
    _N_BYTES_PER_COUNT = BYTES_PER_WORD

    #: size of the increment for spike recording
    _N_BYTES_PER_INCREMENT = BYTES_PER_WORD
    _N_BYTES_PER_SIZE = BYTES_PER_WORD

    # sampling temporal value size (how many ticks between recordings)
    _N_CPU_CYCLES_PER_NEURON = 8
    _N_BYTES_PER_POINTER = BYTES_PER_WORD
    _SARK_BLOCK_SIZE = 2 * BYTES_PER_WORD  # Seen in sark.c

    #: size of the counter for outstanding recording
    _N_BYTES_PER_OUTSTANDING_RECORDING = BYTES_PER_WORD

    #: number of items types (currently non-bitfield and bitfield)
    _N_ITEM_TYPES = 2

    #: flag for spikes
    SPIKES = "spikes"

    #: packets-per-timestep
    PACKETS = "packets-per-timestep"

    #: packets-per-timestep data type
    PACKETS_TYPE = DataType.UINT32

    #: rewiring
    REWIRING = "rewiring"

    #: rewiring data type
    REWIRING_TYPE = DataType.UINT32

    #: max_rewires
    MAX_REWIRES = "max_rewires"

    _MAX_RATE = 2 ** 32 - 1  # To allow a unit32_t to be used to store the rate

    def __init__(
            self, allowed_variables: List[str],
            data_types: Mapping[str, DataType],
            bitfield_variables: Sequence[str],
            n_neurons: int, per_timestep_variables: Sequence[str],
            per_timestep_datatypes: Mapping[str, DataType],
            events_per_core_variables: Sequence[str],
            events_per_core_datatypes: Mapping[str, DataType]):
        """
        :param allowed_variables:
        :param data_types:
        :param bitfield_variables:
        :param n_neurons:
        :param per_timestep_variables:
        :param per_timestep_datatypes:
        :param events_per_core_variables:
        :param events_per_core_datatypes:
        """
        self.__sampling_rates: Dict[str, int] = dict()
        self.__indexes: Dict[str, Optional[Sequence[int]]] = dict()
        self.__data_types = data_types
        self.__n_neurons = n_neurons
        self.__bitfield_variables = bitfield_variables

        self.__per_timestep_variables = per_timestep_variables
        self.__per_timestep_datatypes = per_timestep_datatypes
        self.__per_timestep_recording: Set[str] = set()

        self.__events_per_core_variables = events_per_core_variables
        self.__events_per_core_datatypes = events_per_core_datatypes
        self.__events_per_core_recording: Set[str] = set()
        self.__events_per_ts: Dict[str, int] = dict()
        self.__events_per_ts[self.MAX_REWIRES] = 0  # record('all')

        # Get info on variables like these
        for variable in itertools.chain(allowed_variables, bitfield_variables):
            self.__sampling_rates[variable] = 0
            self.__indexes[variable] = None

        # Get region ids for all variables
        self.__region_ids = {
            variable: region_id
            for region_id, variable in enumerate(itertools.chain(
                allowed_variables, bitfield_variables,
                events_per_core_variables, per_timestep_variables))}

        self.__offset_added = False

    def add_region_offset(self, offset: int) -> None:
        """
        Add an offset to the regions.
        Used when there are multiple recorders on a single core.

        :param offset: The offset to add
        """
        if not self.__offset_added:
            self.__region_ids = dict(
                (var, region + offset)
                for var, region in self.__region_ids.items())

        self.__offset_added = True

    def get_region(self, variable: str) -> int:
        """
        :param variable: The variable to get the region of
        :returns: The region of a variable.
        """
        return self.__region_ids[variable]

    def _rate_and_count_per_slice(
            self, variable: str,
            vertex_slice: Optional[Slice]) -> Tuple[int, int]:
        if variable not in self.__sampling_rates:
            return 0, 0
        if self.__sampling_rates[variable] == 0:
            return 0, 0
        indices = self.__indexes.get(variable)
        if indices is None:
            n_atoms = self.__n_neurons if vertex_slice is None \
                else vertex_slice.n_atoms
            return self.__sampling_rates[variable], n_atoms
        assert vertex_slice is not None
        count = len(numpy.intersect1d(
            vertex_slice.get_raster_ids(), indices))
        if count:
            return self.__sampling_rates[variable], count
        return 0, 0

    def _max_recording_per_slice(
            self, variable: str, n_atoms: int) -> Optional[int]:
        if variable not in self.__sampling_rates:
            return None
        if self.__sampling_rates[variable] == 0:
            return 0
        indices = self.__indexes.get(variable)
        if indices is None:
            return n_atoms
        max_index = numpy.amax(indices)
        existence = numpy.zeros(max_index + 1, dtype=numpy.int32)
        existence[indices] = 1
        splits = numpy.arange(n_atoms, max_index + 1, n_atoms)
        split_array = numpy.array_split(existence, splits)
        return max([int(numpy.sum(s)) for s in split_array])

    def neurons_recording(
            self, variable: str,
            vertex_slice: Slice) -> Optional[Collection[int]]:
        """
        :param variable:
        :param vertex_slice:
        :returns: The indexes of the neurons recording
            or None if the variable is unknown
        """
        if variable not in self.__sampling_rates:
            return None
        if self.__sampling_rates[variable] == 0:
            return []
        indices = self.__indexes.get(variable)
        if indices is None:
            return vertex_slice.get_raster_ids()
        all_set = set(indices)
        ids = set(map(int, vertex_slice.get_raster_ids()))
        return sorted(all_set & ids)

    def _convert_placement_matrix_data(
            self, row_data: NDArray[uint8], n_rows: int, data_row_length: int,
            n_neurons: int, data_type: DataType) -> NDArray[float64]:
        surplus_bytes = self._N_BYTES_FOR_TIMESTAMP
        var_data = row_data[:, surplus_bytes:].reshape(
            n_rows * data_row_length)
        placement_data = data_type.decode_array(var_data).reshape(
            n_rows, n_neurons)
        return placement_data

    @staticmethod
    def _process_missing_data(
            missing_str: str, placement: Placement, expected_rows: int,
            n_neurons: int, times: NDArray[uint32], sampling_rate: int,
            label: str, placement_data: NDArray[float64], region: int
            ) -> NDArray[float64]:
        missing_str += f"({placement.x}, {placement.y}, {placement.p}); "
        # Start the fragment for this slice empty
        fragment = numpy.empty((expected_rows, n_neurons))
        for i in range(expected_rows):
            time = i * sampling_rate
            # Check if there is data for this time step
            local_indexes = numpy.where(times == time)
            if len(local_indexes[0]) == 1:
                fragment[i] = placement_data[local_indexes[0]]
            elif len(local_indexes[0]) > 1:
                fragment[i] = placement_data[local_indexes[0][0]]
                logger.warning(
                    "Population {} has multiple recorded data for time {}"
                    " in region {} ", label, time, region)
            else:
                # Set row to Nan
                fragment[i] = numpy.full(n_neurons, numpy.nan)
        return fragment

    def _get_placement_matrix_data(
            self, vertex: MachineVertex, region: int, expected_rows: int,
            missing_str: str, sampling_rate: int, label: str,
            data_type: DataType, n_per_timestep: int) -> Optional[
                NDArray[float64]]:
        """
        Processes a placement for matrix data.

        :param vertex: the vertex to read from
        :param region: the recording region id
        :param expected_rows:
            how many rows the tools think should be recorded
        :param missing_str: string for reporting missing stuff
        :param sampling_rate: the rate of sampling
        :param  label: the vertex label.
        :return: placement data
        """
        placement = SpynnakerDataView.get_placement_of_vertex(vertex)
        if n_per_timestep == 0:
            return None

        # for buffering output info is taken form the buffer manager
        buffer_manager = SpynnakerDataView.get_buffer_manager()
        record_raw, missing_data = buffer_manager.get_recording(
            placement, region)
        record_length = len(record_raw)

        # If there is no data, return empty for all timesteps
        if record_length == 0:
            return numpy.zeros((expected_rows, n_per_timestep), dtype=float64)

        # There is one column for time and one for each neuron recording
        data_row_length = n_per_timestep * data_type.size
        full_row_length = data_row_length + self._N_BYTES_FOR_TIMESTAMP
        n_rows = record_length // full_row_length
        row_data = numpy.asarray(record_raw, dtype=uint8).reshape(
            n_rows, full_row_length)
        placement_data = self._convert_placement_matrix_data(
            row_data, n_rows, data_row_length, n_per_timestep, data_type)

        # If everything is there, return it
        if not missing_data and n_rows == expected_rows:
            return placement_data

        # Got data but its missing bits, so get times
        time_bytes = (
            row_data[:, 0: self._N_BYTES_FOR_TIMESTAMP].reshape(
                n_rows * self._N_BYTES_FOR_TIMESTAMP))
        times: NDArray[uint32] = time_bytes.view("<i4").reshape(n_rows, 1)

        # process data from core for missing data
        placement_data = self._process_missing_data(
            missing_str, placement, expected_rows, n_per_timestep, times,
            sampling_rate, label, placement_data, region)
        return placement_data

    def get_recorded_indices(
            self, application_vertex: ApplicationVertex,
            variable: str) -> Iterable[int]:
        """
        :param application_vertex: The vertex being recorded
        :param variable: The name of the variable to get the indices of
        :returns: The indices being recorded for a given variable.
        """
        if variable not in self.__sampling_rates:
            return []
        indices = self.__indexes.get(variable)
        if indices is None:
            return range(application_vertex.n_atoms)
        return indices

    def get_sampling_interval_ms(self, variable: str) -> float:
        """
        Get the sampling interval of a variable.

        :param variable: The variable to get the sampling interval of
        :returns: Sampling interval in microseconds
        """
        if (variable in self.__per_timestep_variables or
                variable in self.__events_per_core_variables):
            return get_sampling_interval(1)

        return get_sampling_interval(self.__sampling_rates[variable])

    def get_buffer_data_type(self, variable: str) -> BufferDataType:
        """
        :param variable: Name of variable to get type for
        :returns: Type of data for this variable.
           Defaults to MATRIX even if the variable in unknown
        """
        if variable == self.SPIKES:
            return BufferDataType.NEURON_SPIKES
        elif variable == self.REWIRING:
            return BufferDataType.REWIRES
        elif variable in self.__events_per_core_variables:
            raise NotImplementedError(
                f"Unexpected Event variable: {variable}")
        else:
            return BufferDataType.MATRIX

    def get_data_type(self, variable: str) -> Optional[DataType]:
        """
        :param variable:
        :returns: Type of the Data for this variable or None if unknown
        """
        if variable in self.__per_timestep_variables:
            return self.__per_timestep_datatypes[variable]
        if variable in self.__data_types:
            return self.__data_types[variable]
        return None

    def get_recordable_variables(self) -> List[str]:
        """
        :returns: Names of variables that can be recorded
        """
        return [
            *self.__sampling_rates.keys(),
            *self.__events_per_core_variables,
            *self.__per_timestep_variables]

    def get_event_recordable_variables(self) -> List[str]:
        """
        :returns: Names of event variables being recorded
        """
        return list(self.__events_per_core_variables)

    def is_recording(self, variable: str) -> bool:
        """
        :param variable:
        :returns: True if the variable is being recorded,
           False if the variable is not being recorded or not recordable
        """
        try:
            return self.__sampling_rates[variable] > 0
        except KeyError:
            if (variable in self.__events_per_core_recording or
                    variable in self.__per_timestep_recording):
                return True
        return False

    def is_recordable(self, variable: str) -> bool:
        """
        :param variable: The variable to check for
        :returns: True if the given variable can be recorded,
           regardless if it is actually recording or not.
        """
        return (variable in self.__sampling_rates or
                variable in self.__per_timestep_variables or
                variable in self.__events_per_core_variables)

    @property
    def recording_variables(self) -> Iterable[str]:
        """
        Names of the variables recording
        """
        for variable, rate in self.__sampling_rates.items():
            if rate > 0:
                yield variable
        for variable in self.__events_per_core_variables:
            if variable in self.__events_per_core_recording:
                yield variable
        for variable in self.__per_timestep_variables:
            if variable in self.__per_timestep_recording:
                yield variable

    @property
    def recorded_region_ids(self) -> Iterable[int]:
        """
        The ids of the regions for each variable being recorded.
        """
        for variable, rate in self.__sampling_rates.items():
            if rate > 0:
                yield self.__region_ids[variable]
        # events per core regions come after normal regions
        for variable in self.__events_per_core_variables:
            if variable in self.__events_per_core_recording:
                yield self.__region_ids[variable]
        # Per timestep regions come next
        for variable in self.__per_timestep_variables:
            if variable in self.__per_timestep_recording:
                yield self.__region_ids[variable]

    def _is_recording(self, variable: str, vertex_slice: Slice) -> bool:
        # event per core and per_timestep variables are not recorded by slice,
        # so True if present
        if variable in self.__events_per_core_recording:
            return True
        if variable in self.__per_timestep_recording:
            return True
        if self.__sampling_rates[variable] == 0:
            return False
        indices = self.__indexes.get(variable)
        if indices is None:
            return True
        return any(numpy.isin(vertex_slice.get_raster_ids(), indices))

    def recorded_ids_by_slice(self, vertex_slice: Slice) -> List[int]:
        """
        :param vertex_slice:
        :returns: list of the IDs of the variables actually recording.
           Editing this list will not change recording settings.
        """
        # This needs to produce a new list every time; callers modify it
        variables = [
            self.__region_ids[variable]
            for variable in self.__sampling_rates
            if self._is_recording(variable, vertex_slice)]
        # event per core variables are always defined, but the region is
        # declared after the other variables
        variables.extend(
            self.__region_ids[variable]
            for variable in self.__events_per_core_variables
            if variable in self.__events_per_core_recording)
        # per-timestep variables are always defined, but the region is declared
        # after the other variables
        variables.extend(
            self.__region_ids[variable]
            for variable in self.__per_timestep_variables
            if variable in self.__per_timestep_recording)
        return variables

    def _compute_rate(self, sampling_interval: Optional[float]) -> int:
        """
        Convert a sampling interval into a rate.
        Remember, machine time step is in nanoseconds

        :param sampling_interval: interval between samples in microseconds
        :return: rate
        """
        if sampling_interval is None:
            return 1

        step = SpynnakerDataView.get_simulation_time_step_ms()
        rate = int(sampling_interval / step)
        if sampling_interval != rate * step:
            raise ConfigurationException(
                f"sampling_interval {sampling_interval} is not an an "
                f"integer multiple of the simulation timestep {step}")
        if rate > self._MAX_RATE:
            raise ConfigurationException(
                f"sampling_interval {sampling_interval} higher than "
                f"max allowed which is {step * self._MAX_RATE}")
        return rate

    def _check_indexes(self, indexes: Optional[Collection[int]]) -> None:
        if indexes is None:
            return

        if len(indexes) == 0:
            raise ConfigurationException("Empty indexes list")

        found = False
        warning = True
        for index in indexes:
            if index < 0:
                raise ConfigurationException(
                    "Negative indexes are not supported")
            elif index >= self.__n_neurons:
                if warning:
                    logger.warning(
                        "Ignoring indexes greater than population size")
                warning = False
            else:
                found = True
        if not found:
            raise ConfigurationException(
                "All indexes larger than population size")

    def __check_per_timestep_params(
            self, variable: str, sampling_interval: Optional[float],
            indexes: Optional[Collection[int]]) -> None:
        """
        Check if certain parameters have been provided for a per-timestep
        variable and if so, raise an Exception.
        """
        if sampling_interval is not None:
            raise ValueError(
                f"Variable {variable} does not support a sampling interval")
        if indexes is not None:
            raise ValueError(
                f"Variable {variable} can only be recorded "
                "on the whole population")

    def __check_events_per_core_params(
            self, variable: str, sampling_interval: Optional[float],
            indexes: Optional[Collection[int]]) -> None:
        """
        Check if certain parameters have been provided for an
        events-per-core variable and if so, raise an Exception.
        """
        if sampling_interval is not None:
            raise ValueError(
                f"Variable {variable} does not support a sampling interval")
        if indexes is not None:
            raise ValueError(
                f"Variable {variable} can only be recorded "
                "on the whole population")

    def _turn_off_recording(
            self, variable: str, sampling_interval: Optional[float],
            remove_indexes: Optional[Collection[int]]) -> None:
        # If a per-timestep variable, remove it and return
        if variable in self.__per_timestep_variables:
            if variable in self.__per_timestep_recording:
                self.__per_timestep_recording.remove(variable)
            return

        # If an events-per-core variable, remove it and return
        if variable in self.__events_per_core_variables:
            if variable in self.__events_per_core_recording:
                self.__events_per_core_recording.remove(variable)
            return

        if self.__sampling_rates[variable] == 0:
            # Already off so ignore other parameters
            return

        if remove_indexes is None:
            # turning all off so ignoring sampling interval
            self.__sampling_rates[variable] = 0
            self.__indexes[variable] = None
            return

        # No good reason to specify_interval when turning off
        if sampling_interval is not None:
            rate = self._compute_rate(sampling_interval)
            # But if they do make sure it is the same as before
            if rate != self.__sampling_rates[variable]:
                raise ConfigurationException(
                    "Illegal sampling_interval parameter while turning "
                    "off recording")

        indexes = self.__indexes.get(variable)
        if indexes is None:
            # start with all indexes
            indexes = range(self.__n_neurons)

        # remove the indexes not recording
        indexes = [
            index
            for index in indexes
            if index not in remove_indexes]

        # Check is at least one index still recording
        if len(indexes) == 0:
            self.__sampling_rates[variable] = 0
            self.__indexes[variable] = None
        else:
            self.__indexes[variable] = indexes

    def _check_complete_overwrite(
            self, variable: str, indexes: Optional[Collection[int]]) -> None:
        if indexes is None:
            # overwriting all OK!
            return
        current = self.__indexes.get(variable)
        if current is None:
            if set(range(self.__n_neurons)).issubset(set(indexes)):
                # overwriting all previous so OK!
                return
        else:
            if set(current).issubset(set(indexes)):
                # overwriting all previous so OK!
                return
        raise ConfigurationException(
            "Current implementation does not support multiple "
            f"sampling_intervals for {variable} on one population.")

    def _turn_on_recording(
            self, variable: str, sampling_interval: Optional[float],
            indexes: Optional[Collection[int]]) -> None:
        # If a per-timestep variable, update
        if variable in self.__per_timestep_variables:
            self.__check_per_timestep_params(
                variable, sampling_interval, indexes)
            self.__per_timestep_recording.add(variable)
            return

        # If an events-per-core variable, update
        if variable in self.__events_per_core_variables:
            self.__check_events_per_core_params(
                variable, sampling_interval, indexes)
            self.__events_per_core_recording.add(variable)
            return

        rate = self._compute_rate(sampling_interval)
        if self.__sampling_rates[variable] == 0:
            # Previously not recording so OK
            self.__sampling_rates[variable] = rate
        elif rate != self.__sampling_rates[variable]:
            self._check_complete_overwrite(variable, indexes)
        # else rate not changed so no action

        if indexes is None:
            # previous recording indexes does not matter as now all (None)
            self.__indexes[variable] = None
        else:
            # make sure indexes is not a generator like range
            indices = set(indexes)
            self._check_indexes(indices)
            current = self.__indexes.get(variable)
            if current:
                # merge the two indexes
                indices.update(current)
            # Keep in numerical order
            self.__indexes[variable] = sorted(indices)

    def set_recording(self, variable: str, new_state: bool,
                      sampling_interval: Optional[float] = None,
                      indexes: Optional[Collection[int]] = None) -> None:
        """
        Turns on the requested recording

        :param variable: PyNN variable name
        :param new_state:
        :param sampling_interval:
        :param indexes:
        """
        if variable == "all":
            for key in self.__sampling_rates:
                self.set_recording(key, new_state, sampling_interval, indexes)
            for var in self.__events_per_core_variables:
                # Skip the unsupported items for an events-per-core variable
                self.set_recording(var, new_state)
            for var in self.__per_timestep_variables:
                # Skip the unsupported items for a per-timestep variable
                self.set_recording(var, new_state)
        elif (variable in self.__sampling_rates or
                variable in self.__per_timestep_variables or
                variable in self.__events_per_core_variables):
            if new_state:
                self._turn_on_recording(variable, sampling_interval, indexes)
            else:
                self._turn_off_recording(variable, sampling_interval, indexes)
        else:
            raise ConfigurationException(
                f"Variable {variable} is not supported")

    def get_region_sizes(self, vertex_slice: Slice) -> List[int]:
        """
        Get the sizes of the regions for the variables, whether they are
        recorded or not, with those that are not having a size of 0.

        :param vertex_slice:
        :returns: List of size of each region including ones not recoding
        """
        return [
            self.get_buffered_sdram(variable, vertex_slice)
            for variable in itertools.chain(
                self.__sampling_rates, self.__events_per_core_variables,
                self.__per_timestep_variables)]

    def write_neuron_recording_region(
            self, spec: DataSpecificationBase,
            neuron_recording_region: int, vertex_slice: Slice) -> None:
        """
        Recording data specification.

        :param spec: where to write the data specification
        :param neuron_recording_region: the recording region
        :param vertex_slice: the vertex slice
        """
        spec.switch_write_focus(neuron_recording_region)

        # Write the number of variables and bitfields (ignore per-timestep)
        n_vars = len(self.__sampling_rates) - len(self.__bitfield_variables)
        spec.write_value(data=n_vars)
        spec.write_value(data=len(self.__bitfield_variables))

        # Write the recording data
        recording_data = self._get_data(vertex_slice)
        spec.write_array(recording_data)

    def _get_buffered_sdram_per_record(
            self, variable: str, n_neurons: int) -> int:
        """
        :param variable: PyNN variable name
        :return: The SDRAM used per record.

        """
        if variable in self.__per_timestep_variables:
            if variable not in self.__per_timestep_recording:
                return 0
            size = self.__per_timestep_datatypes[variable].size
            return self._N_BYTES_FOR_TIMESTAMP + size
        if variable in self.__events_per_core_variables:
            if variable not in self.__events_per_core_recording:
                return 0
            size = self.__events_per_core_datatypes[variable].size
            return self.__events_per_ts[self.MAX_REWIRES] * (
                self._N_BYTES_FOR_TIMESTAMP + size)
        if n_neurons == 0:
            return 0
        if variable in self.__bitfield_variables:
            # Overflow can be ignored as it is not save if in an extra word
            out_spike_words = int(math.ceil(n_neurons / BITS_PER_WORD))
            out_spike_bytes = out_spike_words * BYTES_PER_WORD
            return self._N_BYTES_FOR_TIMESTAMP + out_spike_bytes
        else:
            size = self.__data_types[variable].size
            return self._N_BYTES_FOR_TIMESTAMP + (n_neurons * size)

    def get_buffered_sdram_per_record(
            self, variable: str, vertex_slice: Slice) -> int:
        """
        Return the SDRAM used per record.

        :param variable: PyNN variable name
        :param vertex_slice:
        :return: usage
        """
        _, n_neurons = self._rate_and_count_per_slice(variable, vertex_slice)
        return self._get_buffered_sdram_per_record(variable, n_neurons)

    def get_max_buffered_sdram_per_record(
            self, variable: str, n_atoms: int) -> int:
        """
        Return the SDRAM used per record.

        :param variable: PyNN variable name
        :param n_atoms: number of atoms recording
        :return: usage
        """
        n_neurons = self._max_recording_per_slice(variable, n_atoms) or 0
        return self._get_buffered_sdram_per_record(variable, n_neurons)

    def get_buffered_sdram_per_timestep(
            self, variable: str, vertex_slice: Slice) -> int:
        """
        Return the SDRAM used per timestep.

        In the case where sampling is used it returns the average
        for recording and none recording based on the recording rate

        :param variable: PyNN variable name
        :param vertex_slice:
        :return: usage
        """
        if variable in self.__per_timestep_variables:
            if variable not in self.__per_timestep_recording:
                return 0
            rate = 1
        elif variable in self.__events_per_core_variables:
            if variable not in self.__events_per_core_recording:
                return 0
            rate = 1
        else:
            rate = self.__sampling_rates[variable]
            if rate == 0:
                return 0

        data_size = self.get_buffered_sdram_per_record(variable, vertex_slice)
        if rate == 1:
            return data_size
        else:
            return data_size // rate

    def get_sampling_overflow_sdram(self, vertex_slice: Slice) -> int:
        """
        Get the extra SDRAM that should be reserved if using per_timestep.

        This is the extra that must be reserved if per_timestep is an average
        rather than fixed for every timestep.

        When sampling the average * time_steps may not be quite enough.
        This returns the extra space in the worst case
        where time_steps is a multiple of sampling rate + 1,
        and recording is done in the first and last time_step

        :param vertex_slice:
        :return: Highest possible overflow needed
        """
        # No need to consider per-timestep variables here as they won't
        # overflow
        overflow = 0
        for variable, rate in self.__sampling_rates.items():
            # If rate is 0 no recording so no overflow
            # If rate is 1 there is no overflow as average is exact
            if rate > 1:
                data_size = self.get_buffered_sdram_per_record(
                    variable, vertex_slice)
                overflow += data_size // rate * (rate - 1)
        return overflow

    def get_buffered_sdram(
            self, variable: str, vertex_slice: Slice) -> int:
        """
        Returns the SDRAM used for this many time steps for a variable.

        If required the total is rounded up so the space will always fit.

        :param variable: The PyNN variable name to get buffered SDRAM of
        :param vertex_slice:
        :return: data size
        """
        n_machine_time_steps = SpynnakerDataView.get_max_run_time_steps()
        # Per timestep variables can't be done at a specific rate
        if variable in self.__per_timestep_variables:
            item = self.get_buffered_sdram_per_record(variable, vertex_slice)
            return item * n_machine_time_steps

        # Events per core variables depend on the max rewires possible
        # (this is already taken into consideration in per_record calculation)
        if variable in self.__events_per_core_variables:
            item = self.get_buffered_sdram_per_record(variable, vertex_slice)
            return item * n_machine_time_steps

        rate = self.__sampling_rates[variable]
        if rate == 0:
            return 0
        data_size = self.get_buffered_sdram_per_record(variable, vertex_slice)
        records = n_machine_time_steps // rate
        if n_machine_time_steps % rate > 0:
            records = records + 1
        return data_size * records

    def get_metadata_sdram_usage_in_bytes(self, n_atoms: int) -> int:
        """
        :param n_atoms: The number of atoms to record
        :returns: The SDRAM usage of the metadata for recording.
        """
        # This calculates the size of the metadata only; thus no reference to
        # per-timestep variables which have no metadata
        n_indices = self.__ceil_n_indices(n_atoms)
        n_bytes_for_indices = n_indices * self._N_BYTES_PER_INDEX
        var_bytes = (
            (self._N_BYTES_PER_RATE + self._N_BYTES_PER_SIZE +
             self._N_BYTES_PER_ENUM + n_bytes_for_indices) *
            (len(self.__sampling_rates) - len(self.__bitfield_variables)))
        bitfield_bytes = (
            (self._N_BYTES_PER_RATE + self._N_BYTES_PER_SIZE +
             n_bytes_for_indices) *
            len(self.__bitfield_variables))
        return ((self._N_ITEM_TYPES * DataType.UINT32.size) + var_bytes +
                bitfield_bytes)

    def get_generator_sdram_usage_in_bytes(self, n_atoms: int) -> int:
        """
        :param n_atoms: The number of atoms to be recorded
        :returns:
           The SDRAM usage of the generator data for recording metadata.
        """
        n_indices = self.__ceil_n_indices(n_atoms)
        n_bytes_for_indices = n_indices * self._N_BYTES_PER_INDEX
        var_bytes = (
            (self._N_BYTES_PER_RATE + self._N_BYTES_PER_SIZE +
             self._N_BYTES_PER_ENUM + self._N_BYTES_PER_GEN_ITEM +
             n_bytes_for_indices) *
            (len(self.__sampling_rates) - len(self.__bitfield_variables)))
        bitfield_bytes = (
            (self._N_BYTES_PER_RATE + self._N_BYTES_PER_SIZE +
             self._N_BYTES_PER_GEN_ITEM + n_bytes_for_indices) *
            len(self.__bitfield_variables))
        return ((self._N_ITEM_TYPES * DataType.UINT32.size) + var_bytes +
                bitfield_bytes)

    def get_variable_sdram_usage(self, vertex_slice: Slice) -> AbstractSDRAM:
        """
        :param vertex_slice:
        :returns: SDRAM cost including fix cost for saving the recordings
        """
        fixed_sdram = 0.0
        per_timestep_sdram = 0.0
        for variable in self.__sampling_rates:
            rate = self.__sampling_rates[variable]
            if rate > 0:
                fixed_sdram += self._SARK_BLOCK_SIZE
                per_record = self.get_buffered_sdram_per_record(
                    variable, vertex_slice)
                if rate == 1:
                    # Add size for one record as recording every timestep
                    per_timestep_sdram += per_record
                else:
                    # Get the average cost per timestep
                    average_per_timestep = per_record / rate
                    per_timestep_sdram += average_per_timestep
                    # Add the rest once to fixed for worst case
                    fixed_sdram += (per_record - average_per_timestep)
        for variable in self.__per_timestep_recording:
            per_timestep_sdram += self.get_buffered_sdram_per_record(
                variable, vertex_slice)
        for variable in self.__events_per_core_recording:
            per_timestep_sdram += self.get_buffered_sdram_per_record(
                variable, vertex_slice)
        return VariableSDRAM(
            math.ceil(fixed_sdram), math.ceil(per_timestep_sdram))

    def get_max_variable_sdram_usage(self, n_atoms: int) -> AbstractSDRAM:
        """
        :param n_atoms:
        :returns: SDRAM cost EXCLUDING fix cost for saving the recordings
        """
        fixed_sdram = 0.0
        per_timestep_sdram = 0.0
        for variable in self.__sampling_rates:
            rate = self.__sampling_rates[variable]
            # fixed_sdram += self._get_fixed_sdram_usage(n_atoms)
            if rate > 0:
                fixed_sdram += self._SARK_BLOCK_SIZE
                per_record = self.get_max_buffered_sdram_per_record(
                    variable, n_atoms)
                if rate == 1:
                    # Add size for one record as recording every timestep
                    per_timestep_sdram += per_record
                else:
                    # Get the average cost per timestep
                    average_per_timestep = per_record / rate
                    per_timestep_sdram += average_per_timestep
                    # Add the rest once to fixed for worst case
                    fixed_sdram += (per_record - average_per_timestep)
        for variable in self.__per_timestep_recording:
            per_timestep_sdram += self.get_max_buffered_sdram_per_record(
                variable, n_atoms)
        for variable in self.__events_per_core_recording:
            per_timestep_sdram += self.get_max_buffered_sdram_per_record(
                variable, n_atoms)
        return VariableSDRAM(
            math.ceil(fixed_sdram), math.ceil(per_timestep_sdram))

    def __ceil_n_indices(self, n_neurons: int) -> int:
        """
        The number of indices rounded up to a whole number of words.

        :param n_neurons: The number of neurons to account for
        """
        # Assumes that BYTES_PER_WORD is divisible by _N_BYTES_PER_INDEX
        n_bytes = n_neurons * self._N_BYTES_PER_INDEX
        ceil_bytes = int(math.ceil(n_bytes / BYTES_PER_WORD)) * BYTES_PER_WORD
        return ceil_bytes // self._N_BYTES_PER_INDEX

    def __add_indices(
            self, data: List[NDArray[uint32]], variable: str, rate: int,
            n_recording: int, vertex_slice: Slice) -> None:
        n_indices = self.__ceil_n_indices(vertex_slice.n_atoms)
        if rate == 0:
            # Not recording anything so all indices are 0
            data.append(numpy.zeros(n_indices, dtype=uint16).view(uint32))
        elif (indexes := self.__indexes.get(variable)) is None:
            # Recording everything so indices are identity
            data.append(numpy.arange(n_indices, dtype=uint16).view(uint32))
        else:
            local_index = 0
            local_indexes: List[int] = list()
            # Add indices based on the raster ids
            index = 0
            for index in vertex_slice.get_raster_ids():
                if index in indexes:
                    # Recording so write the local index to record to
                    local_indexes.append(local_index)
                    local_index += 1
                else:
                    # Not recording so write to one beyond recording range
                    local_indexes.append(n_recording)
            # Any extra indices should be filled in
            for _ in range(len(local_indexes), n_indices):
                local_indexes.append(n_recording)
            data.append(
                numpy.array(local_indexes, dtype=uint16).view(uint32))

    def _get_data(self, vertex_slice: Slice) -> NDArray[uint32]:
        # There is no data here for per-timestep variables by design
        data: List[NDArray[uint32]] = list()
        for variable in self.__sampling_rates:
            rate, n_recording = self._rate_and_count_per_slice(
                variable, vertex_slice)
            if variable in self.__bitfield_variables:
                data.append(numpy.array([rate, n_recording], dtype=uint32))
            else:
                dtype = self.__data_types[variable]
                data.append(numpy.array(
                    [rate, n_recording, dtype.size], dtype=uint32))
            self.__add_indices(data, variable, rate, n_recording, vertex_slice)

        return numpy.concatenate(data)

    def set_max_rewires_per_ts(self, max_rewires_per_ts: int) -> None:
        """
        Sets the the maximum rewires per timestep

        :param max_rewires_per_ts: the maximum rewires per timestep
        """
        self.__events_per_ts[self.MAX_REWIRES] = max_rewires_per_ts

    @property
    def _indexes(self) -> Mapping:  # for testing only
        return MappingProxyType(self.__indexes)

    @property
    def is_global_generatable(self) -> bool:
        """
        Whether the data for all neurons the same, i.e., all or none of the
        neurons are recorded for all variables.
        """
        for variable in self.__sampling_rates:
            if variable in self.__indexes:
                return False
        return True

    def get_generator_data(
            self, vertex_slice: Optional[Slice] = None) -> NDArray[uint32]:
        """
        :param vertex_slice:
            The slice to generate the data for, or `None` to generate for
            all neurons (assuming all the same, otherwise error)
        :returns: The recorded data as a generatable data set.
        """
        n_vars = len(self.__sampling_rates) - len(self.__bitfield_variables)
        data = [n_vars, len(self.__bitfield_variables)]
        for variable in self.__sampling_rates:
            rate = self.__sampling_rates[variable]
            if variable in self.__bitfield_variables:
                data.append(rate)
            else:
                data.extend([rate, self.__data_types[variable].size])
            if rate == 0:
                data.extend([0, 0])
            else:
                data.extend(self.__get_generator_indices(
                    variable, vertex_slice))
        return numpy.array(data, dtype=uint32)

    def __get_generator_indices(
            self, variable: str,
            vertex_slice: Optional[Slice]) -> Iterable[int]:
        """
        Get the indices of the variables to record in run-length-encoded form.
        """
        index = self.__indexes.get(variable)

        # If there is no index, add that all variables are recorded
        if index is None:
            return [_REPEAT_PER_NEURON, 1,
                    _REPEAT_PER_NEURON_RECORDED | _RECORDED_FLAG]

        assert (vertex_slice is not None)

        # Find slice-relative indices in the index
        indices = numpy.intersect1d(
            vertex_slice.get_raster_ids(), index, return_indices=True)[1]

        # If there is no overlap, nothing is recorded
        if len(indices) == 0:
            return [_REPEAT_PER_NEURON, 1,
                    _REPEAT_PER_NEURON_RECORDED | _NOT_RECORDED_FLAG]

        # Split the indices into consecutive ranges
        ranges = numpy.split(
            indices, numpy.where(numpy.diff(indices) > 1)[0] + 1)

        # Generate a run-length-encoded list
        # Initially there are no items, but this will be updated
        # Also keep track of the number recorded, also 0 initially
        data = [0, 0]
        n_items = 0

        last_index = 0
        n_recorded = 0
        for r in ranges:
            # Add the last non-recorded range
            if r[0] > last_index:
                data.append((r[0] - last_index) | _NOT_RECORDED_FLAG)
                n_items += 1

            # Add this recorded range
            data.append(((r[-1] - r[0]) + 1) | _RECORDED_FLAG)
            n_items += 1
            n_recorded += (r[-1] - r[0]) + 1
            last_index = r[-1] + 1

        # Add the final range if needed
        if last_index < vertex_slice.n_atoms:
            data.append((vertex_slice.n_atoms - last_index) |
                        _NOT_RECORDED_FLAG)
            n_items += 1

        data[0] = n_recorded
        data[1] = n_items
        return numpy.array(data, dtype=uint32)



# Copyright (c) 2015 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Iterable

from numpy import floating
from numpy.typing import NDArray

from spinn_utilities.overrides import overrides

from spinn_front_end_common.interface.ds import DataSpecificationBase
from spinn_front_end_common.utilities.constants import (
    BYTES_PER_SHORT, BYTES_PER_WORD)

from spynnaker.pyNN.data import SpynnakerDataView
from spynnaker.pyNN.models.neuron.plasticity.stdp.common import (
    get_exp_lut_array)
from spynnaker.pyNN.models.neuron.plasticity.stdp.synapse_structure import (
    SynapseStructureWeightOnly)

from .abstract_timing_dependence import AbstractTimingDependence


class TimingDependenceSpikePair(AbstractTimingDependence):
    """
    A basic timing dependence STDP rule.
    """
    __slots__ = (
        "__tau_minus",
        "__tau_minus_data",
        "__tau_plus",
        "__tau_plus_data",
        "__a_plus",
        "__a_minus")
    __PARAM_NAMES = ('tau_plus', 'tau_minus')

    def __init__(
            self, tau_plus: float = 20.0, tau_minus: float = 20.0,
            A_plus: float = 0.01, A_minus: float = 0.01):
        r"""
        :param tau_plus: :math:`\tau_+`
        :param tau_minus: :math:`\tau_-`
        :param A_plus: :math:`A^+`
        :param A_minus: :math:`A^-`
        """
        super().__init__(SynapseStructureWeightOnly())
        self.__tau_plus = tau_plus
        self.__tau_minus = tau_minus
        self.__a_plus = A_plus
        self.__a_minus = A_minus

        # provenance data
        ts = SpynnakerDataView.get_simulation_time_step_ms()
        self.__tau_plus_data = get_exp_lut_array(ts, self.__tau_plus)
        self.__tau_minus_data = get_exp_lut_array(ts, self.__tau_minus)

    @property
    def tau_plus(self) -> float:
        r"""
        :math:`\tau_+`
        """
        return self.__tau_plus

    @property
    def tau_minus(self) -> float:
        r"""
        :math:`\tau_-`
        """
        return self.__tau_minus

    @property
    def A_plus(self) -> float:
        r"""
        :math:`A^+`
        """
        return self.__a_plus

    @A_plus.setter
    def A_plus(self, new_value: float) -> None:
        self.__a_plus = new_value

    @property
    def A_minus(self) -> float:
        r"""
        :math:`A^-`
        """
        return self.__a_minus

    @A_minus.setter
    def A_minus(self, new_value: float) -> None:
        self.__a_minus = new_value

    @overrides(AbstractTimingDependence.is_same_as)
    def is_same_as(
            self, timing_dependence: AbstractTimingDependence) -> bool:
        if not isinstance(timing_dependence, TimingDependenceSpikePair):
            return False
        return (self.__tau_plus == timing_dependence.tau_plus and
                self.__tau_minus == timing_dependence.tau_minus)

    @property
    def vertex_executable_suffix(self) -> str:
        """
        The suffix to be appended to the vertex executable for this rule.
        """
        return "pair"

    @property
    def pre_trace_n_bytes(self) -> int:
        """
        The number of bytes used by the pre-trace of the rule per neuron.
        """
        # Pair rule requires no pre-synaptic trace when only the nearest
        # Neighbours are considered and, a single 16-bit R1 trace
        return BYTES_PER_SHORT

    @overrides(AbstractTimingDependence.get_parameters_sdram_usage_in_bytes)
    def get_parameters_sdram_usage_in_bytes(self) -> int:
        return BYTES_PER_WORD * (len(self.__tau_plus_data) +
                                 len(self.__tau_minus_data))

    @property
    def n_weight_terms(self) -> int:
        """
        The number of weight terms expected by this timing rule.
        """
        return 1

    @overrides(AbstractTimingDependence.write_parameters)
    def write_parameters(
            self, spec: DataSpecificationBase, global_weight_scale: float,
            synapse_weight_scales: NDArray[floating]) -> None:
        # Write lookup tables
        spec.write_array(self.__tau_plus_data)
        spec.write_array(self.__tau_minus_data)

    @overrides(AbstractTimingDependence.get_parameter_names)
    def get_parameter_names(self) -> Iterable[str]:
        return self.__PARAM_NAMES



# Copyright (c) 2017 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from spinn_utilities.overrides import overrides
from spynnaker.pyNN.models.neuron.plasticity.stdp.synapse_structure import (
    AbstractSynapseStructure)


class SynapseStructureWeightAccumulator(AbstractSynapseStructure):
    """
    Structured synapse with accumulated weight
    """
    __slots__ = ()

    @overrides(AbstractSynapseStructure.get_n_half_words_per_connection)
    def get_n_half_words_per_connection(self) -> int:
        return 2

    @overrides(AbstractSynapseStructure.get_weight_half_word)
    def get_weight_half_word(self) -> int:
        return 0



# Copyright (c) 2024 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Optional

from pyNN.random import NumpyRNG

from spinn_utilities.overrides import overrides
from spinn_utilities.ranged import RangeDictionary

from spinn_front_end_common.interface.ds import DataType

from spynnaker.pyNN.models.neuron.implementations import ModelParameter
from spynnaker.pyNN.utilities.struct import Struct
from spynnaker.pyNN.random_distribution import RandomDistribution
from .abstract_threshold_type import AbstractThresholdType

V_THRESH = "v_thresh"
P_THRESH = "p_thresh"
SEED0 = "seed0"
SEED1 = "seed1"
SEED2 = "seed2"
SEED3 = "seed3"


class ThresholdTypeFixedProb(AbstractThresholdType):
    """
    A threshold that spikes with a fixed probability when over a static value.
    """
    __slots__ = ("__v_thresh", "__p_thresh", "__random")

    def __init__(self, v_thresh: ModelParameter, p_thresh: ModelParameter,
                 seed: Optional[int]):

        """
        :param v_thresh: :math:`V_{thresh}`
        :param p_thresh: :math:`P_{thresh}`
        :param seed: Random number generator seed
        """
        super().__init__(
            [Struct([(DataType.S1615, V_THRESH),
                     (DataType.U1616, P_THRESH),
                     (DataType.UINT32, SEED0),
                     (DataType.UINT32, SEED1),
                     (DataType.UINT32, SEED2),
                     (DataType.UINT32, SEED3)])],
            {V_THRESH: "mV",
             P_THRESH: ""})
        self.__v_thresh = v_thresh
        self.__p_thresh = p_thresh
        self.__random = RandomDistribution(
            "uniform", low=0, high=0xFFFFFFFF, rng=NumpyRNG(seed))

    @overrides(AbstractThresholdType.add_parameters)
    def add_parameters(self, parameters: RangeDictionary[float]) -> None:
        parameters[V_THRESH] = self._convert(self.__v_thresh)
        parameters[P_THRESH] = self._convert(self.__p_thresh)

    @overrides(AbstractThresholdType.add_state_variables)
    def add_state_variables(
            self, state_variables: RangeDictionary[float]) -> None:
        state_variables[SEED0] = self.__random
        state_variables[SEED1] = self.__random
        state_variables[SEED2] = self.__random
        state_variables[SEED3] = self.__random

    @property
    def v_thresh(self) -> ModelParameter:
        """
        :math:`V_{thresh}`
        """
        return self.__v_thresh

    @property
    def p_thresh(self) -> ModelParameter:
        """
        :math:`P_{thresh}`
        """
        return self.__p_thresh



# Copyright (c) 2017 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import annotations
from typing import (
    Any, Dict, Iterable, Optional, Tuple, TYPE_CHECKING, Sequence, Union)

import numpy

from pyNN.standardmodels.synapses import StaticSynapse

from spinn_utilities.overrides import overrides

from pacman.model.graphs.application import ApplicationVertex
from pacman.model.graphs.common import Slice

from spynnaker.pyNN.exceptions import SynapticConfigurationException
from spynnaker.pyNN.types import WEIGHTS
from spynnaker.pyNN.utilities.utility_calls import create_mars_kiss_seeds

from .abstract_synapse_dynamics_structural import (
    AbstractSynapseDynamicsStructural)
from .synapse_dynamics_structural_common import (
    DEFAULT_F_REW, DEFAULT_INITIAL_WEIGHT, DEFAULT_INITIAL_DELAY,
    DEFAULT_S_MAX, SynapseDynamicsStructuralCommon as
    _Common)
from .abstract_static_synapse_dynamics import AbstractStaticSynapseDynamics
from .synapse_dynamics_static import SynapseDynamicsStatic
from .synapse_dynamics_stdp import SynapseDynamicsSTDP
from .synapse_dynamics_structural_stdp import SynapseDynamicsStructuralSTDP
from .abstract_synapse_dynamics import AbstractSynapseDynamics
from .synapse_dynamics_weight_changable import SynapseDynamicsWeightChangable
from .synapse_dynamics_weight_changer import SynapseDynamicsWeightChanger

if TYPE_CHECKING:
    from pacman.model.graphs import AbstractVertex
    from pacman.model.graphs.machine import MachineVertex
    from spynnaker.pyNN.models.neural_projections import (
        ProjectionApplicationEdge, SynapseInformation)
    from spynnaker.pyNN.models.neural_projections.connectors import (
        AbstractConnector)
    from spynnaker.pyNN.models.neuron.synapse_dynamics.\
        abstract_synapse_dynamics_structural import (
            InitialDelay)
    from spynnaker.pyNN.models.neuron.structural_plasticity.synaptogenesis.\
        partner_selection.abstract_partner_selection import (
            AbstractPartnerSelection)
    from spynnaker.pyNN.models.neuron.structural_plasticity.synaptogenesis.\
        formation.abstract_formation import (
            AbstractFormation)
    from spynnaker.pyNN.models.neuron.structural_plasticity.synaptogenesis.\
        elimination.abstract_elimination import (
            AbstractElimination)
    from spynnaker.pyNN.models.neuron.synapse_dynamics.types import (
        ConnectionsArray)
    from spynnaker.pyNN.types import DELAYS
    from .synapse_dynamics_structural_common import ConnectionsInfo


class SynapseDynamicsStructuralStatic(SynapseDynamicsStatic, _Common):
    """
    Class that enables synaptic rewiring in the absence of STDP.

    It acts as a wrapper around SynapseDynamicsStatic, meaning that rewiring
    can operate in parallel with static synapses.

    Written by Petrut Bogdan.
    """
    __slots__ = (
        # Frequency of rewiring (Hz)
        "__f_rew",
        # Initial weight assigned to a newly formed connection
        "__initial_weight",
        # Delay assigned to a newly formed connection
        "__initial_delay",
        # Maximum fan-in per target layer neuron
        "__s_max",
        # The seed
        "__seed",
        # Holds initial connectivity as defined via connector
        "__connections",
        # The actual type of weights: static through the simulation or those
        # that can be change through STDP
        "__weight_dynamics",
        # Shared RNG seed to be written on all cores
        "__seeds",
        # The RNG used with the seed that is passed in
        "__rng",
        # The partner selection rule
        "__partner_selection",
        # The formation rule
        "__formation",
        # The elimination rule
        "__elimination",
        "__with_replacement")

    def __init__(
            self, partner_selection: AbstractPartnerSelection,
            formation: AbstractFormation, elimination: AbstractElimination,
            f_rew: float = DEFAULT_F_REW,
            initial_weight: float = DEFAULT_INITIAL_WEIGHT,
            initial_delay: InitialDelay = DEFAULT_INITIAL_DELAY,
            s_max: int = DEFAULT_S_MAX,
            with_replacement: bool = True, seed: Optional[int] = None,
            weight: float = StaticSynapse.default_parameters['weight'],
            delay: Optional[float] = None):
        """
        :param partner_selection: The partner selection rule
        :param formation: The formation rule
        :param elimination: The elimination rule
        :param f_rew: How many rewiring attempts will be done per second.
        :param initial_weight:
            Weight assigned to a newly formed connection
        :param initial_delay:
            Delay assigned to a newly formed connection; a single value means
            a fixed delay value, or a tuple of two values means the delay will
            be chosen at random from a uniform distribution between the given
            values
        :param s_max: Maximum fan-in per target layer neuron
        :param with_replacement:
            If set to True (default), a new synapse can be formed in a
            location where a connection already exists; if False, then it must
            form where no connection already exists
        :param seed: seed the random number generators
        :param weight: The weight of connections formed by the connector
        :param delay: The delay of connections formed by the connector
            Use ``None`` to get the simulator default minimum delay.
        """
        super().__init__(weight=weight, delay=delay, pad_to_length=s_max)

        self.__partner_selection = partner_selection
        self.__formation = formation
        self.__elimination = elimination
        self.__f_rew = f_rew
        self.__initial_weight = initial_weight
        self.__initial_delay = initial_delay
        self.__s_max = s_max
        self.__with_replacement = with_replacement
        self.__seed = seed
        self.__connections: ConnectionsInfo = dict()

        self.__rng = numpy.random.RandomState(seed)
        self.__seeds: Dict[Any, Tuple[int, ...]] = dict()

    @overrides(AbstractStaticSynapseDynamics.merge)
    def merge(self, synapse_dynamics: AbstractSynapseDynamics
              ) -> AbstractSynapseDynamics:
        if isinstance(synapse_dynamics, (SynapseDynamicsWeightChangable,
                                         SynapseDynamicsWeightChanger)):
            raise SynapticConfigurationException(
                "Weight Changer and Structural Plasticity are not currently"
                " compatible")

        # If the dynamics is structural, check if same as this
        if isinstance(synapse_dynamics, AbstractSynapseDynamicsStructural):
            if not self.is_same_as(synapse_dynamics):
                raise SynapticConfigurationException(
                    "Synapse dynamics must match exactly when using multiple"
                    " edges to the same population")

            # If structural part matches, return other as it might also be STDP
            return synapse_dynamics

        # If the dynamics is STDP but not Structural (as here), merge
        if isinstance(synapse_dynamics, SynapseDynamicsSTDP):
            return SynapseDynamicsStructuralSTDP(
                self.partner_selection, self.formation,
                self.elimination,
                synapse_dynamics.timing_dependence,
                synapse_dynamics.weight_dependence,
                # voltage dependence is not supported
                None, synapse_dynamics.dendritic_delay_fraction,
                self.f_rew, self.initial_weight, self.initial_delay,
                self.s_max, self.with_replacement, self.seed,
                backprop_delay=synapse_dynamics.backprop_delay)

        # Otherwise, it is static, so return ourselves
        return self

    @overrides(AbstractStaticSynapseDynamics.is_same_as)
    @overrides(_Common.is_same_as)
    def is_same_as(self, synapse_dynamics: Union[
            AbstractSynapseDynamics,
            AbstractSynapseDynamicsStructural]) -> bool:
        if not (isinstance(synapse_dynamics, SynapseDynamicsStructuralStatic)):
            return False
        if not AbstractStaticSynapseDynamics.is_same_as(
                self, synapse_dynamics):
            return False
        return _Common.is_same_as(self, synapse_dynamics)

    @overrides(AbstractStaticSynapseDynamics.get_vertex_executable_suffix)
    def get_vertex_executable_suffix(self) -> str:
        return (super().get_vertex_executable_suffix() +
                _Common.get_vertex_executable_suffix(self))

    @overrides(AbstractSynapseDynamicsStructural.set_connections)
    def set_connections(
            self, connections: ConnectionsArray, post_vertex_slice: Slice,
            app_edge: ProjectionApplicationEdge,
            synapse_info: SynapseInformation) -> None:
        if not isinstance(synapse_info.synapse_dynamics,
                          AbstractSynapseDynamicsStructural):
            return
        collector = self.__connections.setdefault(
            (app_edge.post_vertex, post_vertex_slice.lo_atom), [])
        collector.append((connections, app_edge, synapse_info))

    @overrides(AbstractStaticSynapseDynamics.get_parameter_names)
    def get_parameter_names(self) -> Iterable[str]:
        yield from super().get_parameter_names()
        yield from _Common.get_parameter_names(self)

    @property
    @overrides(SynapseDynamicsStatic.changes_during_run)
    def changes_during_run(self) -> bool:
        return True

    @property
    @overrides(AbstractSynapseDynamicsStructural.f_rew)
    def f_rew(self) -> float:
        return self.__f_rew

    @property
    @overrides(AbstractSynapseDynamicsStructural.seed)
    def seed(self) -> Optional[int]:
        return self.__seed

    @property
    @overrides(AbstractSynapseDynamicsStructural.s_max)
    def s_max(self) -> int:
        return self.__s_max

    @property
    @overrides(AbstractSynapseDynamicsStructural.with_replacement)
    def with_replacement(self) -> bool:
        return self.__with_replacement

    @property
    @overrides(AbstractSynapseDynamicsStructural.initial_weight)
    def initial_weight(self) -> float:
        return self.__initial_weight

    @property
    @overrides(AbstractSynapseDynamicsStructural.initial_delay)
    def initial_delay(self) -> InitialDelay:
        return self.__initial_delay

    @property
    @overrides(AbstractSynapseDynamicsStructural.partner_selection)
    def partner_selection(self) -> AbstractPartnerSelection:
        return self.__partner_selection

    @property
    @overrides(AbstractSynapseDynamicsStructural.formation)
    def formation(self) -> AbstractFormation:
        return self.__formation

    @property
    @overrides(AbstractSynapseDynamicsStructural.elimination)
    def elimination(self) -> AbstractElimination:
        return self.__elimination

    @property
    @overrides(_Common.connections)
    def connections(self) -> ConnectionsInfo:
        return self.__connections

    @overrides(SynapseDynamicsStatic.get_weight_mean)
    def get_weight_mean(self, connector: AbstractConnector,
                        synapse_info: SynapseInformation) -> float:
        return self.get_weight_maximum(connector, synapse_info)

    @overrides(SynapseDynamicsStatic.get_weight_variance)
    def get_weight_variance(
           self, connector: AbstractConnector, weights: WEIGHTS,
           synapse_info: SynapseInformation) -> float:
        return 0.0

    @overrides(SynapseDynamicsStatic.get_weight_maximum)
    def get_weight_maximum(self, connector: AbstractConnector,
                           synapse_info: SynapseInformation) -> float:
        w_m = super().get_weight_maximum(connector, synapse_info)
        return max(w_m, self.__initial_weight)

    @overrides(SynapseDynamicsStatic.get_delay_maximum)
    def get_delay_maximum(self, connector: AbstractConnector,
                          synapse_info: SynapseInformation) -> float:
        d_m = super().get_delay_maximum(connector, synapse_info)
        if d_m is None:
            return self.__initial_delay
        return max(d_m, self.__initial_delay)

    @overrides(SynapseDynamicsStatic.get_delay_minimum)
    def get_delay_minimum(
            self, connector: AbstractConnector,
            synapse_info: SynapseInformation) -> Optional[float]:
        d_m = super().get_delay_minimum(connector, synapse_info)
        if d_m is None:
            return self.__initial_delay
        return min(d_m, self.__initial_delay)

    @overrides(SynapseDynamicsStatic.get_delay_variance)
    def get_delay_variance(
            self, connector: AbstractConnector, delays: DELAYS,
            synapse_info: SynapseInformation) -> float:
        return 0.0

    @overrides(_Common._get_seeds)
    def _get_seeds(
            self, app_vertex: Union[None, ApplicationVertex, Slice] = None
            ) -> Sequence[int]:
        if app_vertex:
            if app_vertex not in self.__seeds:
                self.__seeds[app_vertex] = (
                    create_mars_kiss_seeds(self.__rng))
            return self.__seeds[app_vertex]
        else:
            return create_mars_kiss_seeds(self.__rng)

    @overrides(SynapseDynamicsStatic.generate_on_machine)
    def generate_on_machine(self) -> bool:
        # Never generate structural connections on the machine
        return False

    @overrides(AbstractSynapseDynamics.get_connected_vertices)
    def get_connected_vertices(
            self, s_info: SynapseInformation, source_vertex: ApplicationVertex,
            target_vertex: ApplicationVertex) -> Sequence[
            Tuple[MachineVertex, Sequence[AbstractVertex]]]:

        # Things change, so assume all connected
        return [(m_vertex, [source_vertex])
                for m_vertex in target_vertex.splitter.get_in_coming_vertices(
                    s_info.partition_id)]

    @property
    @overrides(AbstractSynapseDynamics.is_combined_core_capable)
    def is_combined_core_capable(self) -> bool:
        return False

    @property
    @overrides(AbstractSynapseDynamics.is_split_core_capable)
    def is_split_core_capable(self) -> bool:
        return True

    @property
    @overrides(AbstractStaticSynapseDynamics.synapses_per_second)
    def synapses_per_second(self) -> int:
        return (super().synapses_per_second -
                super().reduction_synapses_per_second)



# Copyright (c) 2021 The University of Manchester
# Based on work Copyright (c) The University of Sussex,
# Garibaldi Pineda Garcia, James Turner, James Knight and Thomas Nowotny
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations
from collections.abc import Iterable, Sized
from typing import (
    Optional, Tuple, Union, cast, TYPE_CHECKING)

import numpy
from numpy import integer, floating, float64, uint16, uint32
from numpy.typing import ArrayLike, NDArray

from pyNN.random import RandomDistribution

from spinn_utilities.overrides import overrides

from pacman.model.graphs.common import Slice

from spinn_front_end_common.utilities.constants import (
    BYTES_PER_WORD, BYTES_PER_SHORT)
from spinn_front_end_common.utilities.exceptions import ConfigurationException

from spynnaker.pyNN.exceptions import SynapticConfigurationException
from spynnaker.pyNN.models.common.local_only_2d_common import get_div_const

from .abstract_connector import AbstractConnector

if TYPE_CHECKING:
    from spynnaker.pyNN.models.neural_projections import (
        ProjectionApplicationEdge, SynapseInformation)


_DIMENSION_SIZE = BYTES_PER_WORD
_CONN_SIZE = (6 * BYTES_PER_SHORT)


class PoolDenseConnector(AbstractConnector):
    """
    A multidimensional connector based on weight and pool shape and stride.

    A weight is provided for every pre-post pair of neurons,
    where the post-population size is determined by the pre-population size
    and a pool shape,
    which how many pre-neurons are connected to each post-neuron
    in each dimension, and pool stride, which decides how the shape moves over
    the image in each dimension to choose the pre-neurons for each post-neuron.
    """

    __slots__ = (
        "__weights",
        "__pool_shape",
        "__pool_stride",
        "__positive_receptor_type",
        "__negative_receptor_type")

    def __init__(self, weights: ArrayLike,
                 pool_shape: Union[int, Tuple[int], None] = None,
                 pool_stride: Union[int, Tuple[int], None] = None,
                 positive_receptor_type: str = "excitatory",
                 negative_receptor_type: str = "inhibitory",
                 safe: bool = True, verbose: bool = False,
                 callback: None = None):
        """
        :param weights:
            The synaptic strengths. Can be:

            * single value: the same value will be used for all weights
            * :py:class:`list`: the total number of elements must be
              (number after pooling * number post)
            * :py:class:`~numpy.ndarray`: As above for list
            * :py:class:`~spynnaker.pyNN.RandomDistribution`:
              weights will be drawn at random
        :param pool_shape:
            Shape of average pooling. If a single value is provided, it will
            be used for every dimension, otherwise must be the same number of
            values as there are dimensions in the source.
        :param pool_stride:
            Jumps between pooling regions. If a single value is provided, the
            same stride will be used for all dimensions, otherwise must be
            the same number of values as there are dimensions in the source.
            If `None`, and pool_shape is provided, pool_stride will be set to
            pool_shape.
        :param positive_receptor_type:
            The receptor type to add the positive weights to.  By default this
            is "excitatory".
        :param negative_receptor_type:
            The receptor type to add the negative weights to.  By default this
            is "inhibitory".
        :param safe: (ignored)
        :param verbose: (ignored)
        :param callback:
            if given, a callable that display a progress bar on the terminal.

            .. note::
                Not supported by sPyNNaker.
        """
        super().__init__(safe=safe, callback=callback, verbose=verbose)
        self.__weights = numpy.array(weights)
        self.__pool_shape = pool_shape
        self.__pool_stride = pool_shape if pool_stride is None else pool_stride
        self.__positive_receptor_type = positive_receptor_type
        self.__negative_receptor_type = negative_receptor_type

    @property
    def positive_receptor_type(self) -> str:
        """
        The receptor type to add the positive weights to
        """
        return self.__positive_receptor_type

    @property
    def negative_receptor_type(self) -> str:
        """
        The receptor type to add the negative weights to
        """
        return self.__negative_receptor_type

    @property
    def weights(self) -> NDArray:
        """
        The synaptic strengths
        """
        return self.__weights

    def __decode_weights(
            self, pre_shape: Tuple[int, ...], post_shape: Tuple[int, ...],
            post_vertex_slice: Slice) -> NDArray[float64]:
        if isinstance(self.__weights, (int, float)):
            n_weights = self.__get_n_weights(
                pre_shape, post_vertex_slice.n_atoms)
            return numpy.full(n_weights, self.__weights, dtype=float64)
        elif isinstance(self.__weights, Iterable):
            pre_in_post_shape = tuple(self.__get_pre_in_post_shape(pre_shape))
            all_weights = numpy.array(self.__weights, dtype=float64).reshape(
                pre_in_post_shape + post_shape)
            pip_slices = tuple(
                slice(0, pip_end + 1) for pip_end in pre_in_post_shape)
            # TODO check this is correct
            post_slices = post_vertex_slice.dimension
            return all_weights[pip_slices + post_slices].flatten()
        elif isinstance(self.__weights, RandomDistribution):
            n_weights = self.__get_n_weights(
                pre_shape, post_vertex_slice.n_atoms)
            # pylint: disable=no-member
            # see https://github.com/SpiNNakerManchester/sPyNNaker/issues/1436
            return numpy.array(self.__weights.next(n_weights), dtype=float64)
        else:
            raise SynapticConfigurationException(
                f"Unknown weights ({self.__weights})")

    @staticmethod
    def __to_nd_shape_or_none(
            shape: Optional[Union[int, Tuple[int, ...]]], n_dims: int,
            param_name: str) -> Optional[NDArray[integer]]:
        if shape is None:
            return None
        return PoolDenseConnector.__to_nd_shape(shape, n_dims, param_name)

    @staticmethod
    def __to_nd_shape(shape: Union[int, Tuple[int, ...]],
                      n_dims: int, param_name: str) -> NDArray[integer]:
        if numpy.isscalar(shape):
            return numpy.array([shape] * n_dims, dtype=int)
        shape_tuple = cast(Sized, shape)
        if len(shape_tuple) == n_dims:
            return numpy.array(shape_tuple, dtype=int)
        raise SynapticConfigurationException(
            f"{param_name} must be an int or a tuple(int) with {n_dims}"
            " dimensions")

    @classmethod
    def get_post_pool_shape(
            cls, pre_shape: Tuple[int, ...],
            pool_shape: Union[int, Tuple[int, ...], None] = None,
            pool_stride: Union[int, Tuple[int, ...], None] = None) -> NDArray:
        """
        :param pre_shape: tuple(int)
        :param pool_shape:
        :param pool_stride:
        :return: The shape considering the stride
        """
        real_pool_shape = cls.__to_nd_shape_or_none(
            pool_shape, len(pre_shape), "pool_shape")
        real_pool_stride = cls.__to_nd_shape_or_none(
            pool_stride, len(pre_shape), "pool_stride")
        if real_pool_stride is None:
            real_pool_stride = real_pool_shape
        shape = numpy.array(pre_shape)
        if real_pool_shape is not None:
            shape = shape // real_pool_stride
        return shape

    def __get_pre_in_post_shape(self, pre_shape: Tuple[int, ...]) -> NDArray:
        return self.get_post_pool_shape(
            pre_shape, self.__pool_shape, self.__pool_stride)

    def __get_n_weights(
            self, pre_shape: Tuple[int, ...],
            post_n_atoms: int) -> int:
        """
        Get the expected number of weights.
        """
        shape = self.__get_pre_in_post_shape(pre_shape)
        return numpy.prod(shape) * post_n_atoms

    @overrides(AbstractConnector.validate_connection)
    def validate_connection(
            self, application_edge: ProjectionApplicationEdge,
            synapse_info: SynapseInformation) -> None:
        pre = application_edge.pre_vertex
        post = application_edge.post_vertex
        if len(pre.atoms_shape) != 2:
            raise ConfigurationException(
                "The PoolDenseConnector only works where the pre-Population"
                " of a Projection is 2D.  Please ensure that the"
                " Population uses a Grid2D structure.")

        if isinstance(self.__weights, Iterable):
            expected_n_weights = self.__get_n_weights(
                pre.atoms_shape, post.n_atoms)
            if expected_n_weights != numpy.array(self.__weights).size:
                raise ConfigurationException(
                    f"With a source population with shape {pre.atoms_shape},"
                    f" and a target population with shape {post.atoms_shape},"
                    f" this connector requires {expected_n_weights} weights")
        if post.get_synapse_id_by_target(
                self.__positive_receptor_type) is None:
            raise ConfigurationException(
                "The post population doesn't have a synaptic receptor type of"
                f" {self.__positive_receptor_type}")
        if post.get_synapse_id_by_target(
                self.__negative_receptor_type) is None:
            raise ConfigurationException(
                "The post population doesn't have a synaptic receptor type of"
                f" {self.__negative_receptor_type}")
        if not isinstance(synapse_info.delays, float):
            raise ConfigurationException(
                "The PoolDenseConnector only supports simple uniform delays")

    @staticmethod
    def __delay(synapse_info: SynapseInformation) -> float:
        # Checked by validate_connection above
        return cast(float, synapse_info.delays)

    @overrides(AbstractConnector.get_delay_minimum)
    def get_delay_minimum(self, synapse_info: SynapseInformation) -> float:
        return self.__delay(synapse_info)

    @overrides(AbstractConnector.get_delay_maximum)
    def get_delay_maximum(self, synapse_info: SynapseInformation) -> float:
        return self.__delay(synapse_info)

    @overrides(AbstractConnector.get_n_connections_from_pre_vertex_maximum)
    def get_n_connections_from_pre_vertex_maximum(
            self, n_post_atoms: int, synapse_info: SynapseInformation,
            min_delay: Optional[float] = None,
            max_delay: Optional[float] = None) -> int:
        if min_delay is not None and max_delay is not None:
            if not (min_delay <= self.__delay(synapse_info) <= max_delay):
                return 0
        # Every pre connects to every post
        return n_post_atoms

    @overrides(AbstractConnector.get_n_connections_to_post_vertex_maximum)
    def get_n_connections_to_post_vertex_maximum(
            self, synapse_info: SynapseInformation) -> int:
        # Every post connects to every pre
        return synapse_info.n_pre_neurons

    @overrides(AbstractConnector.get_weight_maximum)
    def get_weight_maximum(self, synapse_info: SynapseInformation) -> float:
        if isinstance(self.__weights, Iterable):
            return numpy.amax(numpy.abs(self.__weights))
        n_conns = synapse_info.n_pre_neurons * synapse_info.n_post_neurons
        return super()._get_weight_maximum(
            self.__weights, n_conns, synapse_info)

    def local_only_n_bytes(self, pre_shape: Tuple[int, ...],
                           n_post_atoms: int) -> int:
        """
        :param pre_shape:
        :param n_post_atoms:
        :returns: Size in bytes needed for the weights
        """
        n_weights = self.__get_n_weights(pre_shape, n_post_atoms)
        n_weights = n_weights + 1 if n_weights % 2 != 0 else n_weights
        n_dims = len(pre_shape)

        return int((n_dims * _DIMENSION_SIZE) + (n_weights * BYTES_PER_SHORT) +
                   _CONN_SIZE)

    def get_local_only_data(
            self, app_edge: ProjectionApplicationEdge, local_delay: int,
            delay_stage: int, post_vertex_slice: Slice,
            weight_scales: NDArray[floating]) -> NDArray[uint32]:
        """
        :param app_edge:
        :param local_delay:
        :param delay_stage:
        :param post_vertex_slice:
        :param weight_scales:
        :return:
        """

        # Write numbers of things
        n_dims = len(app_edge.pre_vertex.atoms_shape)
        n_weights = self.__get_n_weights(
            app_edge.pre_vertex.atoms_shape, post_vertex_slice.n_atoms)

        # Write synapse information
        pos_synapse_type = app_edge.post_vertex.get_synapse_id_by_target(
            self.__positive_receptor_type)
        neg_synapse_type = app_edge.post_vertex.get_synapse_id_by_target(
            self.__negative_receptor_type)

        short_data = numpy.array([
            n_dims, n_weights, pos_synapse_type, neg_synapse_type,
            delay_stage, local_delay], dtype=numpy.int16).view(uint32)
        all_data = [short_data]

        # Generate the stride information
        if self.__pool_stride is not None:
            stride = self.__to_nd_shape(self.__pool_stride, n_dims, "")
            all_data.append(numpy.array(
                [get_div_const(s) for s in stride], dtype=uint32))
        else:
            all_data.append(numpy.array(
                [get_div_const(1) for _ in range(n_dims)], dtype=uint32))

        # Work out which weights are for this connection
        weights = self.__decode_weights(
            app_edge.pre_vertex.atoms_shape, app_edge.post_vertex.atoms_shape,
            post_vertex_slice)

        # Divide weights by pooling area if needed
        if self.__pool_shape is not None:
            shape = self.__to_nd_shape(self.__pool_shape, n_dims, "")
            area = numpy.prod(shape)
            weights = weights / area

        # Encode weights with weight scaling
        if len(weights) % 2 != 0:
            weights = numpy.concatenate((weights, numpy.zeros(1)))
        neg_weights = weights < 0
        pos_weights = weights > 0
        weights[neg_weights] *= weight_scales[neg_synapse_type]
        weights[pos_weights] *= weight_scales[pos_synapse_type]
        all_data.append(numpy.round(weights).astype(uint16).view(uint32))
        return numpy.concatenate(all_data)
